#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Richard Taylor

"""
Testing framework for performing a variety of unttests for GRAMPS.
"""

import logging

import os
import sys
import unittest
from optparse import OptionParser

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
os.environ["GRAMPS_RESOURCES"] = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), ".."
)


def make_parser():
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option(
        "-v",
        "--verbosity",
        type="int",
        dest="verbose_level",
        default=0,
        help="Level of verboseness",
    )
    parser.add_option(
        "-p",
        "--performance",
        action="store_true",
        dest="performance",
        default=False,
        help="Run the performance tests.",
    )
    return parser


def getTestSuites():
    # Sorry about this line, but it is the easiest way of doing it.
    # It just walks the filetree from '.' downwards and returns
    # a tuple per directory of (dirpath,filelist) if the directory
    # contains any test files.

    paths = [
        (f[0], f[2])
        for f in os.walk(".")
        if len([i for i in f[2] if i[-8:] == "_Test.py"])
    ]

    test_suites = []
    perf_suites = []

    for dir, test_modules in paths:
        sys.path.append(dir)

        for module in test_modules:
            if module[-8:] != "_Test.py":
                continue
            mod = __import__(module[:-3])
            test_suites.append(mod.testSuite())
            try:
                perf_suites.append(mod.perfSuite())
            except:
                pass

    return (test_suites, perf_suites)


def allTheTests():
    return unittest.TestSuite(getTestSuites()[0])


def perfTests():
    return unittest.TestSuite(getTestSuites()[1])


if __name__ == "__main__":
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    console.setFormatter(logging.Formatter("%(name)-12s: %(levelname)-8s %(message)s"))

    logger = logging.getLogger("Gramps")
    logger.addHandler(console)

    (options, args) = make_parser().parse_args()

    if options.verbose_level == 1:
        logger.setLevel(logging.INFO)
        console.setLevel(logging.INFO)
    elif options.verbose_level == 2:
        logger.setLevel(logging.DEBUG)
        console.setLevel(logging.DEBUG)
    elif options.verbose_level == 3:
        logger.setLevel(logging.NOTSET)
        console.setLevel(logging.NOTSET)
    elif options.verbose_level >= 4:
        logger.setLevel(logging.NOTSET)
        console.setLevel(logging.NOTSET)
        os.environ["GRAMPS_SIGNAL"] = "1"
    else:
        logger.setLevel(logging.ERROR)
        console.setLevel(logging.ERROR)

    tests = perfTests() if options.performance else allTheTests()
    unittest.TextTestRunner(verbosity=options.verbose_level).run(tests)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman,
# largely based on the MediaView and SelectPerson by Don Allingham

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
import gc

# -------------------------------------------------------------------------
#
# GTK+
#
# -------------------------------------------------------------------------
from gi.repository import GdkPixbuf, Gtk
from gi.repository import GObject
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.const import THUMBSCALE
from gramps.gen.utils.file import media_path_full
from gramps.gen.utils.thumbnails import get_thumbnail_image
from ..views.treemodels import MediaModel
from .baseselector import BaseSelector
from gramps.gen.const import URL_MANUAL_SECT1

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# SelectObject
#
# -------------------------------------------------------------------------
class SelectObject(BaseSelector):
    def get_window_title(self):
        return _("Select Media Object")

    def get_model_class(self):
        return MediaModel

    def get_from_handle_func(self):
        return self.db.get_media_from_handle

    def get_column_titles(self):
        return [
            (_("Title"), 350, BaseSelector.TEXT, 0),
            (_("ID"), 75, BaseSelector.TEXT, 1),
            (_("Type"), 75, BaseSelector.TEXT, 2),
            (_("Last Change"), 150, BaseSelector.TEXT, 7),
            (_("Path"), 150, BaseSelector.TEXT, 3),
        ]

    def _local_init(self):
        """
        Perform local initialisation for this class
        """
        self.setup_configs("interface.media-sel", 600, 450)

        # insert a ScrolledWindow containing an IconView to display thumbnails
        # of the selected media

        # pixels to pad the image
        padding = 6

        self.iconmodel = Gtk.ListStore(GdkPixbuf.Pixbuf, GObject.TYPE_STRING, object)
        self.iconlist = Gtk.IconView()
        self.track_ref_for_deletion("iconlist")
        self.iconlist.set_pixbuf_column(0)
        self.iconlist.set_item_width(int(THUMBSCALE) + padding * 2)

        text_renderer = Gtk.CellRendererText()
        text_renderer.set_property("wrap-mode", Pango.WrapMode.WORD_CHAR)
        text_renderer.set_property("wrap-width", THUMBSCALE)
        text_renderer.set_property("alignment", Pango.Alignment.CENTER)
        self.iconlist.pack_end(text_renderer, True)
        self.iconlist.add_attribute(text_renderer, "text", 1)

        self.iconlist.set_margin(padding)
        self.iconlist.set_column_spacing(padding)
        self.iconlist.set_model(self.iconmodel)

        # create the scrolled window
        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.set_min_content_height(int(THUMBSCALE) + padding * 2)
        scroll.add(self.iconlist)

        vbox = self.glade.get_object("select_person_vbox")
        vbox.pack_start(scroll, False, True, 0)
        vbox.reorder_child(scroll, 1)
        scroll.show()
        self.selection.connect("changed", self._row_change)

    def _row_change(self, obj):
        self.iconmodel.clear()

        id_list = self.get_selected_ids()
        for handle in self.get_selected_ids():
            if handle:
                obj = self.get_from_handle_func()(handle)
                pix = get_thumbnail_image(media_path_full(self.db, obj.get_path()))
                self.iconmodel.append(row=(pix, obj.get_description(), obj))
        gc.collect()

    def get_config_name(self):
        return __name__

    WIKI_HELP_PAGE = URL_MANUAL_SECT1
    WIKI_HELP_SEC = _("Select_Media_Object_selector", "manual")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2022       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A thumbnailer that uses other system thumbnailers already installed on a
Gmone desktop.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import glob
import logging

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import THUMBSCALE, THUMBSCALE_LARGE, SIZE_LARGE
from gramps.gen.plug import Thumbnailer
from gramps.gen.constfunc import get_env_var, lin

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".thumbnail")


class GnomeThumb(Thumbnailer):
    def __init__(self):
        self.all_mime_types = {}
        if lin():
            self.__find_thumbnailers()

    def __find_thumbnailers(self):
        path_list = GLib.get_system_data_dirs()
        path_list.append(GLib.get_user_data_dir())
        file_list = []
        for path in path_list:
            file_list.extend(
                glob.glob(os.path.join(path, "thumbnailers", "*.thumbnailer"))
            )

        for filename in file_list:
            kf = GLib.KeyFile()
            try:
                kf.load_from_file(filename, GLib.KeyFileFlags.NONE)
                cmd = kf.get_string("Thumbnailer Entry", "Exec")
                mime_types = kf.get_string_list("Thumbnailer Entry", "MimeType")
            except GLib.GError:
                continue
            if cmd and mime_types:
                for mime_type in mime_types:
                    self.all_mime_types[mime_type] = cmd

    def is_supported(self, mime_type):
        return mime_type in self.all_mime_types

    def run(self, mime_type, src_file, dest_file, size, rectangle):
        cmd = self.all_mime_types.get(mime_type)
        if cmd:
            if size == SIZE_LARGE:
                thumbscale = THUMBSCALE_LARGE
            else:
                thumbscale = THUMBSCALE
            sublist = {
                "%s": "%d" % int(thumbscale),
                "%u": "file://" + src_file,
                "%i": src_file,
                "%o": dest_file,
            }
            cmdlist = [sublist.get(x, x) for x in cmd.split()]
            return os.spawnvpe(os.P_WAIT, cmdlist[0], cmdlist, os.environ) == 0
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011-2016  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman,
# largely based on the BaseDoc classes by Don Allingham

"""the non-UI-specific (i.e. common, shared) classes for books"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import copy
import os

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".Book")

# -------------------------------------------------------------------------
#
# SAX interface
#
# -------------------------------------------------------------------------
from xml.sax import make_parser, handler, SAXParseException
from xml.sax.saxutils import escape

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ...const import USER_DATA
from ...utils.cast import get_type_converter_by_name, type_name
from ..docgen import StyleSheet, StyleSheetList
from .. import BasePluginManager
from . import book_categories

# ------------------------------------------------------------------------
#
# Private Constants
#
# ------------------------------------------------------------------------
_UNSUPPORTED = _("Unsupported")


# ------------------------------------------------------------------------
#
# Book Item class
#
# ------------------------------------------------------------------------
class BookItem:
    """
    Interface into the book item -- a smallest element of the book.
    """

    def __init__(self, dbase, name):
        """
        Create a new empty BookItem.
        TODO: it should be possible to make a non-empty BookItem, a copy

        name:   the book item is retrieved
                from the book item registry using name for lookup
        """
        self.dbase = dbase
        self.style_name = "default"
        pmgr = BasePluginManager.get_instance()

        for pdata in pmgr.get_reg_bookitems():
            if pdata.id == name:
                self.translated_name = pdata.name
                if not pdata.supported:
                    self.category = _UNSUPPORTED
                else:
                    self.category = book_categories[pdata.category]
                mod = pmgr.load_plugin(pdata)
                self.write_item = getattr(mod, pdata.reportclass)
                self.name = pdata.id
                oclass = getattr(mod, pdata.optionclass)
                self.option_class = oclass(self.name, self.dbase)
                self.option_class.load_previous_values()

    def get_name(self):
        """
        Return the name of the item.
        """
        return self.name

    def get_translated_name(self):
        """
        Return the translated name of the item.
        """
        return self.translated_name

    def get_category(self):
        """
        Return the category of the item.
        """
        return self.category

    def get_write_item(self):
        """
        Return the report-writing function of the item.
        """
        return self.write_item

    def set_style_name(self, style_name):
        """
        Set the style name for the item.

        style_name: name of the style to set.
        """
        self.style_name = style_name

    def get_style_name(self):
        """
        Return the style name of the item.
        """
        return self.style_name


# ------------------------------------------------------------------------
#
# Book class
#
# ------------------------------------------------------------------------
class Book:
    """
    Interface into the user-defined Book -- a collection of book items.
    """

    def __init__(self, obj=None, exact_copy=True):
        """
        Create a new empty Book.

        @param obj:         if not None, creates the Book from obj, from the
                            items in obj, instead of creating an empty Book.
        @type obj:          a :class:`.Book` instance
        @param exact_copy:  if True (and obj is not None) the exact same
                            BookItem objects will be in the new Book;
                            if False (and obj is not None) the same number
                            and same type of BookItem objects will be created
        @type exact_copy:   boolean
        """

        self.name = ""  # this is tested for, in several places
        self.dbname = ""
        self.paper_name = None
        self.paper_orientation = None
        self.paper_metric = None
        self.paper_custom_size = None
        self.paper_margins = None
        self.paper_format = None
        self.paper_output = None
        self.item_list = []
        if obj:
            if exact_copy:
                self.item_list = obj.item_list
            else:
                for item in obj.get_item_list():
                    new_item = BookItem(item.dbase, item.get_name())
                    orig_opt_dict = item.option_class.handler.options_dict
                    new_opt_dict = new_item.option_class.handler.options_dict
                    menu = new_item.option_class.menu
                    for optname in orig_opt_dict:
                        new_opt_dict[optname] = orig_opt_dict[optname]
                        menu_option = menu.get_option_by_name(optname)
                        if menu_option:
                            menu_option.set_value(new_opt_dict[optname])
                    new_item.set_style_name(item.get_style_name())
                    self.item_list.append(new_item)

    def set_name(self, name):
        """
        Set the name of the book.

        name:   the name to set.
        """
        self.name = name

    def get_name(self):
        """
        Return the name of the book.
        """
        return self.name

    def get_dbname(self):
        """
        Return the name of the database file used for the book.
        """
        return self.dbname

    def set_dbname(self, name):
        """
        Set the name of the database file used for the book.

        name:   a filename to set.
        """
        self.dbname = name

    def clear(self):
        """
        Clears the contents of the book.
        """
        self.item_list = []

    def append_item(self, item):
        """
        Add an item to the book.

        item:   an item to append.
        """
        self.item_list.append(item)

    def insert_item(self, index, item):
        """
        Inserts an item into the given position in the book.

        index:  a position index.
        item:   an item to append.
        """
        self.item_list.insert(index, item)

    def pop_item(self, index):
        """
        Pop an item from given position in the book.

        index:  a position index.
        """
        return self.item_list.pop(index)

    def get_item(self, index):
        """
        Return an item at a given position in the book.

        index:  a position index.
        """
        return self.item_list[index]

    def set_item(self, index, item):
        """
        Set an item at a given position in the book.

        index:  a position index.
        item:   an item to set.
        """
        self.item_list[index] = item

    def get_item_list(self):
        """
        Return list of items in the current book.
        """
        return self.item_list

    def set_paper_name(self, paper_name):
        """
        Set the paper name for the Book.
        @param paper_name: name of the paper to set.
        @type paper_name: str
        """
        self.paper_name = paper_name

    def get_paper_name(self):
        """
        Return the paper name of the Book.
        @returns: returns the paper name
        @rtype: str
        """
        return self.paper_name

    def set_orientation(self, orientation):
        """
        Set the paper orientation for the Book.
        @param orientation: orientation to set. Possible values are
            PAPER_LANDSCAPE or PAPER_PORTRAIT
        @type orientation: int
        """
        self.paper_orientation = orientation

    def get_orientation(self):
        """
        Return the paper orientation for the Book.
        @returns: returns the selected orientation. Valid values are
            PAPER_LANDSCAPE or PAPER_PORTRAIT
        @rtype: int
        """
        return self.paper_orientation

    def set_paper_metric(self, paper_metric):
        """
        Set the paper metric for the Book.
        @param paper_metric: whether to use metric.
        @type paper_metric: boolean
        """
        self.paper_metric = paper_metric

    def get_paper_metric(self):
        """
        Return the paper metric of the Book.
        @returns: returns whether to use metric
        @rtype: boolean
        """
        return self.paper_metric

    def set_custom_paper_size(self, paper_size):
        """
        Set the custom paper size for the Book.
        @param paper_size: paper size to set in cm.
        @type paper_size: [float, float]
        """
        self.paper_custom_size = paper_size

    def get_custom_paper_size(self):
        """
        Return the custom paper size for the Book.
        @returns: returns the custom paper size in cm
        @rtype: [float, float]
        """
        return self.paper_custom_size

    def set_margins(self, margins):
        """
        Set the paper margins for the Book.
        @param margins: margins to set. Possible values are floats in cm
        @type margins: [float, float, float, float]
        """
        self.paper_margins = copy.copy(margins)

    def get_margins(self):
        """
        Return the paper margins for the Book.
        @returns margins: returns the margins, floats in cm
        @rtype margins: [float, float, float, float]
        """
        return copy.copy(self.paper_margins)

    def set_margin(self, pos, value):
        """
        Set a paper margin for the Book.
        @param pos: Position of margin [left, right, top, bottom]
        @param value: floating point in cm
        @type pos: int
        @type value: float
        """
        self.paper_margins[pos] = value

    def get_margin(self, pos):
        """
        Return a paper margin for the Book.
        @param pos: Position of margin [left, right, top, bottom]
        @type pos: int
        @returns: float cm of margin
        @rtype: float
        """
        return self.paper_margins[pos]

    def set_format_name(self, format_name):
        """
        Set the format name for the Book.
        @param format_name: name of the format to set.
        @type format_name: str
        """
        self.paper_format = format_name

    def get_format_name(self):
        """
        Return the format name of the Book.
        @returns: returns the format name
        @rtype: str
        """
        return self.paper_format

    def set_output(self, output):
        """
        Set the output for the Book.
        @param output: name of the output to set.
        @type output: str
        """
        self.paper_output = output

    def get_output(self):
        """
        Return the output of the Book.
        @returns: returns the output name
        @rtype: str
        """
        return self.paper_output


# ------------------------------------------------------------------------
#
# BookList class
#
# ------------------------------------------------------------------------
class BookList:
    """
    Interface into the user-defined list of books.

    BookList is loaded from a specified XML file if it exists.
    """

    def __init__(self, filename, dbase):
        """
        Create a new BookList from the books that may be defined in the
        specified file.

        file:   XML file that contains book items definitions
        """
        self.dbase = dbase
        self.bookmap = {}
        self._needs_saving = None
        self.file = os.path.join(USER_DATA, filename)
        self.parse()

    def delete_book(self, name):
        """
        Remove a book from the list. Since each book must have a
        unique name, the name is used to delete the book.

        name:   name of the book to delete
        """
        del self.bookmap[name]

    ## 2/2016 the string "get_book_map" appears nowhere else in gramps
    ##    def get_book_map(self):
    ##        """
    ##        Return the map of names to books.
    ##        """
    ##        return self.bookmap
    ##
    def get_book(self, name):
        """
        Return the Book associated with the name

        name:   name associated with the desired Book.
        """
        return self.bookmap[name]

    def get_book_names(self):
        "Return a list of all the book names in the BookList, sorted"
        return sorted(self.bookmap.keys())

    def set_book(self, name, book):
        """
        Add or replaces a Book in the BookList.

        name:   name associated with the Book to add or replace.
        book:   definition of the book -- a :class:`.Book` instance
        """
        self.bookmap[name] = book

    def set_needs_saving(self, needs_saving):
        """
        Set the needs_saving flag for the BookList.

        @param needs_saving: whether the current BookList needs saving
        @type needs_saving: boolean
        """
        self._needs_saving = needs_saving

    def get_needs_saving(self):
        """
        Return the needs_saving flag of the BookList.

        @returns: returns whether the current BookList needs saving to a file
        @rtype: boolean
        """
        return self._needs_saving

    def save(self):
        """
        Saves the current BookList to the associated file.
        """
        with open(self.file, "w", encoding="utf-8") as b_f:
            b_f.write('<?xml version="1.0" encoding="utf-8"?>\n')
            b_f.write("<booklist>\n")
            for name in sorted(self.bookmap):  # enable a diff of archived copies
                book = self.get_book(name)
                dbname = escape(book.get_dbname())
                b_f.write(
                    '  <book name="%s" database="%s">' "\n" % (escape(name), dbname)
                )
                for item in book.get_item_list():
                    b_f.write(
                        '    <item name="%s" '
                        'trans_name="%s">\n'
                        % (item.get_name(), item.get_translated_name())
                    )
                    options = item.option_class.handler.options_dict
                    for option_name in sorted(options.keys()):  # enable a diff
                        option_value = options[option_name]
                        if isinstance(option_value, (list, tuple)):
                            b_f.write(
                                '      <option name="%s" value="" '
                                'length="%d">\n'
                                % (escape(option_name), len(options[option_name]))
                            )
                            for list_index, v in enumerate(option_value):
                                option_type = type_name(v)
                                value = escape(str(v))
                                value = value.replace('"', "&quot;")
                                b_f.write(
                                    '        <listitem number="%d" '
                                    'type="%s" value="%s"/>\n'
                                    % (list_index, option_type, value)
                                )
                            b_f.write("      </option>\n")
                        else:
                            option_type = type_name(option_value)
                            value = escape(str(option_value))
                            value = value.replace('"', "&quot;")
                            b_f.write(
                                '      <option name="%s" type="%s" '
                                'value="%s"/>\n'
                                % (escape(option_name), option_type, value)
                            )

                    b_f.write('      <style name="%s"/>' "\n" % item.get_style_name())
                    b_f.write("    </item>\n")
                if book.get_paper_name():
                    b_f.write('    <paper name="%s"/>' "\n" % book.get_paper_name())
                if book.get_orientation() is not None:  # 0 is legal
                    b_f.write(
                        '    <orientation value="%s"/>' "\n" % book.get_orientation()
                    )
                if book.get_paper_metric() is not None:  # 0 is legal
                    b_p_metric = book.get_paper_metric()
                    if isinstance(b_p_metric, bool):
                        b_p_metric = int(b_p_metric)
                    b_f.write('    <metric value="%s"/>' "\n" % b_p_metric)
                if book.get_custom_paper_size():
                    size = book.get_custom_paper_size()
                    b_f.write('    <size value="%f %f"/>' "\n" % (size[0], size[1]))
                if book.get_margins():
                    for pos, margin in enumerate(book.get_margins()):
                        b_f.write(
                            '    <margin number="%s" '
                            'value="%f"/>\n' % (pos, book.get_margin(pos))
                        )
                if book.get_format_name():
                    b_f.write('    <format name="%s"/>' "\n" % book.get_format_name())
                if book.get_output():
                    b_f.write(
                        '    <output name="%s"/>' "\n" % escape(book.get_output())
                    )
                b_f.write("  </book>\n")

            b_f.write("</booklist>\n")

    def parse(self):
        """
        Loads the BookList from the associated file, if it exists.
        """
        try:
            parser = make_parser()
            parser.setContentHandler(BookParser(self, self.dbase))
            # bug 10387; XML should be utf8, but was not previously saved
            # that way.  So try to read utf8, if fails, try with system
            # encoding.  Only an issue on non-utf8 systems.
            try:
                with open(self.file, encoding="utf-8") as the_file:
                    parser.parse(the_file)
            except UnicodeDecodeError:
                with open(self.file) as the_file:
                    parser.parse(the_file)
        except (
            IOError,
            OSError,
            ValueError,
            SAXParseException,
            KeyError,
            AttributeError,
        ):
            LOG.debug("Failed to parse book list", exc_info=True)


# -------------------------------------------------------------------------
#
# BookParser
#
# -------------------------------------------------------------------------
class BookParser(handler.ContentHandler):
    """
    SAX parsing class for the Books XML file.
    """

    def __init__(self, booklist, dbase):
        """
        Create a BookParser class that populates the passed booklist.

        booklist:   BookList to be loaded from the file.
        """
        handler.ContentHandler.__init__(self)
        self.dbase = dbase
        self.booklist = booklist
        self.book = None
        self.item = None
        self.option = None
        self.an_opt_name = None
        self.an_opt_value = None
        self.style = None
        self.bname = None
        self.iname = None
        self.dbname = None
        self.b_p_name = None
        self.b_p_orient = None
        self.b_p_metric = None
        self.b_p_size = None
        self.b_p_margins = None
        self.b_p_format = None
        self.b_p_output = None

    def startElement(self, tag, attrs):
        """
        Overridden class that handles the start of a XML element
        """
        if tag == "book":
            self.book = Book()
            self.bname = attrs["name"]
            self.book.set_name(self.bname)
            self.dbname = attrs["database"]
            self.book.set_dbname(self.dbname)
            self.b_p_name = None
            self.b_p_orient = None
            self.b_p_metric = None
            self.b_p_size = None
            self.b_p_margins = None
            self.b_p_format = None
            self.b_p_output = None
        elif tag == "item":
            self.item = BookItem(self.dbase, attrs["name"])
            self.option = {}
        elif tag == "option":
            self.an_opt_name = attrs["name"]
            if "length" in attrs:
                self.an_opt_value = []
            else:
                converter = get_type_converter_by_name(attrs["type"])
                self.an_opt_value = converter(attrs["value"])
        elif tag == "listitem":
            converter = get_type_converter_by_name(attrs["type"])
            self.an_opt_value.append(converter(attrs["value"]))
        elif tag == "style":
            self.style = attrs["name"]
        elif tag == "paper":
            self.b_p_name = attrs["name"]
        elif tag == "orientation":
            self.b_p_orient = int(attrs["value"])
        elif tag == "metric":
            self.b_p_metric = int(attrs["value"])
        elif tag == "size":
            width, height = attrs["value"].split()
            self.b_p_size = [float(width), float(height)]
        elif tag == "margin":
            if self.b_p_margins is None:
                self.b_p_margins = [0.0, 0.0, 0.0, 0.0]
            self.b_p_margins[int(attrs["number"])] = float(attrs["value"])
        elif tag == "format":
            self.b_p_format = attrs["name"]
        elif tag == "output":
            self.b_p_output = attrs["name"]
        else:
            pass

    def endElement(self, tag):
        """
        Overridden class that handles the end of a XML element
        """
        if tag == "option":
            self.option[self.an_opt_name] = self.an_opt_value
        elif tag == "item":
            self.item.option_class.handler.options_dict.update(self.option)
            self.item.set_style_name(self.style)
            self.book.append_item(self.item)
        elif tag == "book":
            if self.b_p_name:
                self.book.set_paper_name(self.b_p_name)
            if self.b_p_orient is not None:  # 0 is legal
                self.book.set_orientation(self.b_p_orient)
            if self.b_p_metric is not None:  # 0 is legal
                self.book.set_paper_metric(self.b_p_metric)
            if self.b_p_size:
                self.book.set_custom_paper_size(self.b_p_size)
            if self.b_p_margins:
                self.book.set_margins(self.b_p_margins)
            if self.b_p_format:
                self.book.set_format_name(self.b_p_format)
            if self.b_p_output:
                self.book.set_output(self.b_p_output)
            self.booklist.set_book(self.bname, self.book)


# -------------------------------------------------------------------------
#
# Functions
#
# -------------------------------------------------------------------------
def append_styles(selected_style, item):
    """
    Append the styles for a book item to the stylesheet.
    """
    ihandler = item.option_class.handler

    # Set up default style
    ihandler.set_default_stylesheet_name(item.get_style_name())
    default_style = StyleSheet()
    make_default_style = item.option_class.make_default_style
    make_default_style(default_style)

    # Read all style sheets available for this item
    style_file = ihandler.get_stylesheet_savefile()
    style_list = StyleSheetList(style_file, default_style)

    # Get the selected stylesheet
    style_name = ihandler.get_default_stylesheet_name()
    style_sheet = style_list.get_style_sheet(style_name)

    for this_style_name in style_sheet.get_paragraph_style_names():
        selected_style.add_paragraph_style(
            this_style_name, style_sheet.get_paragraph_style(this_style_name)
        )

    for this_style_name in style_sheet.get_draw_style_names():
        selected_style.add_draw_style(
            this_style_name, style_sheet.get_draw_style(this_style_name)
        )

    for this_style_name in style_sheet.get_table_style_names():
        selected_style.add_table_style(
            this_style_name, style_sheet.get_table_style(this_style_name)
        )

    for this_style_name in style_sheet.get_cell_style_names():
        selected_style.add_cell_style(
            this_style_name, style_sheet.get_cell_style(this_style_name)
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
#               2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..views.treemodels import PeopleBaseModel, PersonTreeModel
from .baseselector import BaseSelector
from gramps.gen.const import URL_MANUAL_SECT1

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------


# -------------------------------------------------------------------------
#
# SelectEvent
#
# -------------------------------------------------------------------------
class SelectPerson(BaseSelector):
    def __init__(
        self,
        dbstate,
        uistate,
        track=[],
        title=None,
        search_or_filter=None,
        skip=set(),
        show_search_bar=True,
        default=None,
    ):
        # SelectPerson may have a title passed to it which should be used
        # instead of the default defined for get_window_title()
        if title is not None:
            self.title = title
        self.WIKI_HELP_PAGE = URL_MANUAL_SECT1
        if title == _("Select Father"):
            self.WIKI_HELP_SEC = _("Select_Father_selector", "manual")
        elif title == _("Select Mother"):
            self.WIKI_HELP_SEC = _("Select_Mother_selector", "manual")
        elif title == _("Select Child"):
            self.WIKI_HELP_SEC = _("Select_Child_selector", "manual")
        else:
            self.WIKI_HELP_SEC = _("Select_Person_selector", "manual")

        BaseSelector.__init__(
            self,
            dbstate,
            uistate,
            track,
            search_or_filter,
            skip,
            show_search_bar,
            default,
        )

    def _local_init(self):
        """
        Perform local initialisation for this class
        """
        self.setup_configs("interface.person-sel", 600, 450)
        self.tree.connect("key-press-event", self._key_press)

    def get_window_title(self):
        return _("Select Person")

    def get_model_class(self):
        return PersonTreeModel

    def get_column_titles(self):
        return [
            (_("Name"), 250, BaseSelector.TEXT, 0),
            (_("ID"), 75, BaseSelector.TEXT, 1),
            (_("Gender"), 75, BaseSelector.TEXT, 2),
            (_("Birth Date"), 150, BaseSelector.MARKUP, 3),
            (_("Birth Place"), 150, BaseSelector.MARKUP, 4),
            (_("Death Date"), 150, BaseSelector.MARKUP, 5),
            (_("Death Place"), 150, BaseSelector.MARKUP, 6),
            (_("Spouse"), 150, BaseSelector.TEXT, 7),
            (_("Last Change"), 150, BaseSelector.TEXT, 14),
        ]

    def get_from_handle_func(self):
        return self.db.get_person_from_handle

    def exact_search(self):
        """
        Returns a tuple indicating columns requiring an exact search
        """
        return (2,)  # Gender ('female' contains the string 'male')

    def _on_row_activated(self, treeview, path, view_col):
        store, paths = self.selection.get_selected_rows()
        if paths and len(paths[0].get_indices()) == 2:
            self.window.response(Gtk.ResponseType.OK)

    def _key_press(self, obj, event):
        if event.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            store, paths = self.selection.get_selected_rows()
            if paths and len(paths[0].get_indices()) == 1:
                if self.tree.row_expanded(paths[0]):
                    self.tree.collapse_row(paths[0])
                else:
                    self.tree.expand_row(paths[0], 0)
                return True
        return False

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

""""""

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------


# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.errors import ReportError
from gramps.gen.plug.menu import PersonOption
from gramps.gen.plug.docgen import (
    IndexMark,
    FontStyle,
    ParagraphStyle,
    TableStyle,
    TableCellStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_CENTER,
    INDEX_TYPE_TOC,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.simple import SimpleAccess


# ------------------------------------------------------------------------
#
# NoteLinkReport
#
# ------------------------------------------------------------------------
class NoteLinkReport(Report):
    """
    This report
    """

    def write_report(self):
        """
        The routine that actually creates the report.
        At this point, the document is opened and ready for writing.
        """
        sdb = SimpleAccess(self.database)

        self.doc.start_paragraph("NoteLink-Title")
        title = _("Note Link Check Report")
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()
        self.doc.start_table("NoteLinkTable", "NoteLink-Table")

        self.doc.start_row()

        self.doc.start_cell("NoteLink-TableCell")
        self.doc.start_paragraph("NoteLink-Normal-Bold")
        self.doc.write_text(_("Note ID"))
        self.doc.end_paragraph()
        self.doc.end_cell()

        self.doc.start_cell("NoteLink-TableCell")
        self.doc.start_paragraph("NoteLink-Normal-Bold")
        self.doc.write_text(_("Link Type"))
        self.doc.end_paragraph()
        self.doc.end_cell()

        self.doc.start_cell("NoteLink-TableCell")
        self.doc.start_paragraph("NoteLink-Normal-Bold")
        self.doc.write_text(_("Links To"))
        self.doc.end_paragraph()
        self.doc.end_cell()

        self.doc.start_cell("NoteLink-TableCell")
        self.doc.start_paragraph("NoteLink-Normal-Bold")
        self.doc.write_text(_("Status"))
        self.doc.end_paragraph()
        self.doc.end_cell()

        self.doc.end_row()

        if self._user:
            self._user.begin_progress(
                _("Note Link Check Report"),
                _("Generating report"),
                self.database.get_number_of_notes(),
            )
        for note in self.database.iter_notes():
            if self._user:
                self._user.step_progress()
            for ldomain, ltype, lprop, lvalue in note.get_links():
                if ldomain == "gramps":
                    tagtype = _(ltype)
                    ref_obj = sdb.get_link(ltype, lprop, lvalue)
                    if ref_obj:
                        tagvalue = sdb.describe(ref_obj)
                        tagcheck = _("Ok")
                    else:
                        tagvalue = "%s://%s/%s/%s" % (ldomain, ltype, lprop, lvalue)
                        tagcheck = _("Failed")
                else:
                    tagtype = _("Internet")
                    tagvalue = lvalue
                    tagcheck = ""

                self.doc.start_row()

                self.doc.start_cell("NoteLink-TableCell")
                self.doc.start_paragraph("NoteLink-Normal")
                self.doc.write_text(note.gramps_id)
                self.doc.end_paragraph()
                self.doc.end_cell()

                self.doc.start_cell("NoteLink-TableCell")
                self.doc.start_paragraph("NoteLink-Normal")
                self.doc.write_text(tagtype)
                self.doc.end_paragraph()
                self.doc.end_cell()

                self.doc.start_cell("NoteLink-TableCell")
                self.doc.start_paragraph("NoteLink-Normal")
                self.doc.write_text(tagvalue)
                self.doc.end_paragraph()
                self.doc.end_cell()

                self.doc.start_cell("NoteLink-TableCell")
                self.doc.start_paragraph("NoteLink-Normal")
                self.doc.write_text(tagcheck)
                self.doc.end_paragraph()
                self.doc.end_cell()

                self.doc.end_row()
        if self._user:
            self._user.end_progress()

        self.doc.end_table()


# ------------------------------------------------------------------------
#
# NoteLinkOptions
#
# ------------------------------------------------------------------------
class NoteLinkOptions(MenuReportOptions):
    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return _("Entire Database")

    def add_menu_options(self, menu):
        """
        Add options to the menu for the tag report.
        """
        pass

    def make_default_style(self, default_style):
        """Make the default output style for the Note Link Report."""
        # Paragraph Styles
        f = FontStyle()
        f.set_size(16)
        f.set_type_face(FONT_SANS_SERIF)
        f.set_bold(1)
        p = ParagraphStyle()
        p.set_header_level(1)
        p.set_bottom_border(1)
        p.set_top_margin(utils.pt2cm(3))
        p.set_bottom_margin(utils.pt2cm(3))
        p.set_font(f)
        p.set_alignment(PARA_ALIGN_CENTER)
        p.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("NoteLink-Title", p)

        font = FontStyle()
        font.set(face=FONT_SANS_SERIF, size=14, italic=1)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_header_level(2)
        para.set_top_margin(0.25)
        para.set_bottom_margin(0.25)
        para.set_description(_("The style used for the section headers."))
        default_style.add_paragraph_style("NoteLink-Heading", para)

        font = FontStyle()
        font.set_size(12)
        p = ParagraphStyle()
        p.set(first_indent=-0.75, lmargin=0.75)
        p.set_font(font)
        p.set_top_margin(utils.pt2cm(3))
        p.set_bottom_margin(utils.pt2cm(3))
        p.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("NoteLink-Normal", p)

        font = FontStyle()
        font.set_size(12)
        font.set_bold(True)
        p = ParagraphStyle()
        p.set(first_indent=-0.75, lmargin=0.75)
        p.set_font(font)
        p.set_top_margin(utils.pt2cm(3))
        p.set_bottom_margin(utils.pt2cm(3))
        p.set_description(_("The basic style used for table headings."))
        default_style.add_paragraph_style("NoteLink-Normal-Bold", p)

        # Table Styles
        cell = TableCellStyle()
        default_style.add_cell_style("NoteLink-TableCell", cell)

        table = TableStyle()
        table.set_width(100)
        table.set_columns(4)
        table.set_column_width(0, 10)
        table.set_column_width(1, 15)
        table.set_column_width(2, 65)
        table.set_column_width(3, 10)
        default_style.add_table_style("NoteLink-Table", table)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for a person who has a particular relationship.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....lib.familyreltype import FamilyRelType
from .. import Rule

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasRelationship
#
# -------------------------------------------------------------------------
class HasRelationship(Rule):
    """
    Rule that checks for a person who has a particular relationship.
    """

    labels = [
        _("Number of relationships:"),
        _("Relationship type:"),
        _("Number of children:"),
    ]
    name = _("People with the <relationships>")
    description = _("Matches people with a particular relationship")
    category = _("Family filters")

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.relationship_type = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things we only want to do once.
        """
        if self.list[1]:
            self.relationship_type = FamilyRelType()
            self.relationship_type.set_from_xml_str(self.list[1])

    def apply_to_one(self, db: Database, obj: Person) -> bool:
        """
        Apply the rule. Return True on a match.
        """
        relationship_type = 0
        total_children = 0
        number_relations = len(obj.family_list)

        # count children and look for a relationship type match
        for handle in obj.family_list:
            family = db.get_family_from_handle(handle)
            if family:
                total_children += len(family.child_ref_list)
                if self.relationship_type and (self.relationship_type == family.type):
                    relationship_type = 1

        # if number of relations specified
        if self.list[0]:
            try:
                if int(self.list[0]) != number_relations:
                    return False
            except:
                return False

        # number of children
        if self.list[2]:
            try:
                if int(self.list[2]) != total_children:
                    return False
            except:
                return False

        # relation
        if self.list[1]:
            return relationship_type == 1

        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017      Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for JSON schema"""

import json
import os
import unittest

import jsonschema

from ...const import DATA_DIR
from ...db.utils import import_as_dict
from ...user import User
from .. import (
    Citation,
    Event,
    Family,
    Media,
    Note,
    Person,
    Place,
    Repository,
    Source,
    Tag,
)
from ..json_utils import object_to_dict

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")


class BaseTest(unittest.TestCase):
    def _schema_test(self, obj):
        instance = object_to_dict(obj)
        try:
            jsonschema.validate(instance, self.schema)
        except jsonschema.exceptions.ValidationError:
            self.fail("JSON Schema validation error")


class PersonTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Person.get_schema()


class FamilyTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Family.get_schema()


class EventTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Event.get_schema()


class PlaceTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Place.get_schema()


class RepositoryTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Repository.get_schema()


class SourceTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Source.get_schema()


class CitationTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Citation.get_schema()


class MediaTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Media.get_schema()


class NoteTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Note.get_schema()


class TagTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        cls.schema = Tag.get_schema()


def generate_case(obj, test_class):
    """
    Dynamically generate tests.
    """

    def test(self):
        self._schema_test(obj)

    name = "test_schema_%s_%s" % (obj.__class__.__name__, obj.handle)
    setattr(test_class, name, test)


db = import_as_dict(EXAMPLE, User())
for obj in db.iter_people():
    generate_case(obj, PersonTest)
for obj in db.iter_families():
    generate_case(obj, FamilyTest)
for obj in db.iter_events():
    generate_case(obj, EventTest)
for obj in db.iter_places():
    generate_case(obj, PlaceTest)
for obj in db.iter_repositories():
    generate_case(obj, RepositoryTest)
for obj in db.iter_sources():
    generate_case(obj, SourceTest)
for obj in db.iter_citations():
    generate_case(obj, CitationTest)
for obj in db.iter_media():
    generate_case(obj, MediaTest)
for obj in db.iter_notes():
    generate_case(obj, NoteTest)
for obj in db.iter_tags():
    generate_case(obj, TagTest)

if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
try:
    set()
except NameError:
    from sets import Set as set

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import List, Set
from ....lib import Person
from ....db import Database
from ....types import PersonHandle


# -------------------------------------------------------------------------
#
# IsDescendantFamilyOf
#
# -------------------------------------------------------------------------
class IsDescendantFamilyOf(Rule):
    """Rule that checks for a person that is a descendant or the spouse
    of a descendant of a specified person"""

    labels = [_("ID:"), _("Inclusive:")]
    name = _("Descendant family members of <person>")
    category = _("Descendant filters")
    description = _(
        "Matches people that are descendants or the spouse "
        "of a descendant of a specified person"
    )

    def prepare(self, db: Database, user):
        self.db = db
        self.selected_handles: Set[PersonHandle] = set()
        self.root_person = db.get_person_from_gramps_id(self.list[0])
        self.add_matches(self.root_person)
        try:
            if int(self.list[1]):
                inclusive = True
            else:
                inclusive = False
        except IndexError:
            inclusive = True
        if not inclusive:
            self.exclude()

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def add_matches(self, person: Person | None):
        if not person:
            return

        # Add self
        queue: List[Person] = [person]

        while queue:
            person = queue.pop(0)
            if person is None or person.handle in self.selected_handles:
                # if we have been here before, skip
                continue
            self.selected_handles.add(person.handle)
            for family_handle in person.family_list:
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    # Add every child recursively
                    for child_ref in family.child_ref_list:
                        if child_ref:
                            queue.append(self.db.get_person_from_handle(child_ref.ref))
                    # Add spouse
                    if person.handle == family.father_handle:
                        spouse_handle = family.mother_handle
                    else:
                        spouse_handle = family.father_handle
                    self.selected_handles.add(spouse_handle)

    def exclude(self):
        # This removes root person and his/her spouses from the matches set
        if not self.root_person:
            return
        self.selected_handles.remove(self.root_person.handle)
        for family_handle in self.root_person.family_list:
            family = self.db.get_family_from_handle(family_handle)
            if family:
                if self.root_person.handle == family.father_handle:
                    spouse_handle = family.mother_handle
                else:
                    spouse_handle = family.father_handle
                self.selected_handles.remove(spouse_handle)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Martin Hawlisch, Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2022       Jan Skarvall
# Copyright (C) 2022       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Validate localized date parser and displayer.

Based on the Check Localized Date Displayer and Parser tool.
"""

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import unittest

import sys

if "-v" in sys.argv or "--verbose" in sys.argv:
    import logging

    logging.getLogger("").addHandler(logging.StreamHandler())
    log = logging.getLogger(".Date")
    log.setLevel(logging.DEBUG)

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...config import config
from ...lib import Date, DateError
from ...utils.grampslocale import GrampsLocale, _LOCALE_NAMES
from .. import LANG_TO_PARSER


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class DateHandlerTest(unittest.TestCase):
    def setUp(self):
        config.set("preferences.date-format", 0)

    def __base_test_all_languages(self, dates):
        languages = [
            lang for lang in LANG_TO_PARSER.keys() if lang in _LOCALE_NAMES.keys()
        ]
        for language in languages:
            with self.subTest(lang=language):
                self.__test_language(language, dates)

    def __test_language(self, language, dates):
        locale = GrampsLocale(lang=language)
        displayer = locale.date_displayer
        parser = locale.date_parser
        for test_date in dates:
            datestr = displayer.display(test_date)
            new_date = parser.parse(datestr)
            with self.subTest(date=datestr):
                self.assertTrue(
                    test_date.is_equal(new_date),
                    "{} -> {}\n{} -> {}".format(
                        test_date, new_date, test_date.__dict__, new_date.__dict__
                    ),
                )

    def test_simple(self):
        dates = []
        for calendar in (Date.CAL_GREGORIAN, Date.CAL_JULIAN):
            for newyear in (Date.NEWYEAR_JAN1, Date.NEWYEAR_MAR25, (5, 5)):
                for quality in (
                    Date.QUAL_NONE,
                    Date.QUAL_ESTIMATED,
                    Date.QUAL_CALCULATED,
                ):
                    for modifier in (
                        Date.MOD_NONE,
                        Date.MOD_BEFORE,
                        Date.MOD_AFTER,
                        Date.MOD_ABOUT,
                        Date.MOD_FROM,
                        Date.MOD_TO,
                    ):
                        for slash1 in (False, True):
                            for month in (2, 6, 12):
                                for day in (5, 27):
                                    d = Date()
                                    d.set(
                                        quality,
                                        modifier,
                                        calendar,
                                        (day, month, 1789, slash1),
                                        "Text comment",
                                        newyear,
                                    )
                                    dates.append(d)
        self.__base_test_all_languages(dates)

    def test_span(self):
        dates = []
        calendar = Date.CAL_GREGORIAN
        for quality in (Date.QUAL_NONE, Date.QUAL_ESTIMATED, Date.QUAL_CALCULATED):
            for modifier in (Date.MOD_RANGE, Date.MOD_SPAN):
                for slash1 in (False, True):
                    for slash2 in (False, True):
                        for month in (2, 6, 12):
                            for day in (5, 27):
                                d = Date()
                                d.set(
                                    quality,
                                    modifier,
                                    calendar,
                                    (
                                        day,
                                        month,
                                        1789,
                                        slash1,
                                        day,
                                        month,
                                        1876,
                                        slash2,
                                    ),
                                    "Text comment",
                                )
                                dates.append(d)
                                d = Date()
                                d.set(
                                    quality,
                                    modifier,
                                    calendar,
                                    (
                                        day,
                                        month,
                                        1789,
                                        slash1,
                                        day,
                                        13 - month,
                                        1876,
                                        slash2,
                                    ),
                                    "Text comment",
                                )
                                dates.append(d)
                                d = Date()
                                d.set(
                                    quality,
                                    modifier,
                                    calendar,
                                    (
                                        day,
                                        month,
                                        1789,
                                        slash1,
                                        32 - day,
                                        month,
                                        1876,
                                        slash2,
                                    ),
                                    "Text comment",
                                )
                                dates.append(d)
                                d = Date()
                                d.set(
                                    quality,
                                    modifier,
                                    calendar,
                                    (
                                        day,
                                        month,
                                        1789,
                                        slash1,
                                        32 - day,
                                        13 - month,
                                        1876,
                                        slash2,
                                    ),
                                    "Text comment",
                                )
                                dates.append(d)
        self.__base_test_all_languages(dates)

    def test_textual(self):
        dates = []
        calendar = Date.CAL_GREGORIAN
        modifier = Date.MOD_TEXTONLY
        for quality in (Date.QUAL_NONE, Date.QUAL_ESTIMATED, Date.QUAL_CALCULATED):
            test_date = Date()
            test_date.set(
                quality, modifier, calendar, Date.EMPTY, "This is a textual date"
            )
            dates.append(test_date)
        self.__base_test_all_languages(dates)

    def test_too_few_arguments(self):
        dateval = (4, 7, 1789, False)
        for l in range(1, len(dateval)):
            d = Date()
            with self.assertRaises(DateError):
                d.set(
                    Date.QUAL_NONE,
                    Date.MOD_NONE,
                    Date.CAL_GREGORIAN,
                    dateval[:l],
                    "Text comment",
                )

    def test_too_few_span_arguments(self):
        dateval = (4, 7, 1789, False, 5, 8, 1876, False)
        for l in range(1, len(dateval)):
            d = Date()
            with self.assertRaises(DateError):
                d.set(
                    Date.QUAL_NONE,
                    Date.MOD_SPAN,
                    Date.CAL_GREGORIAN,
                    dateval[:l],
                    "Text comment",
                )

    def test_invalid_day(self):
        d = Date()
        with self.assertRaises(DateError):
            d.set(
                Date.QUAL_NONE,
                Date.MOD_NONE,
                Date.CAL_GREGORIAN,
                (44, 7, 1789, False),
                "Text comment",
            )

    def test_invalid_month(self):
        d = Date()
        with self.assertRaises(DateError):
            d.set(
                Date.QUAL_NONE,
                Date.MOD_NONE,
                Date.CAL_GREGORIAN,
                (4, 77, 1789, False),
                "Text comment",
            )

    def test_invalid_month_with_ny(self):
        d = Date()
        with self.assertRaises(DateError):
            d.set(
                Date.QUAL_NONE,
                Date.MOD_NONE,
                Date.CAL_GREGORIAN,
                (4, 77, 1789, False),
                "Text comment",
                newyear=2,
            )

    def test_invalid_span_day(self):
        d = Date()
        with self.assertRaises(DateError):
            d.set(
                Date.QUAL_NONE,
                Date.MOD_SPAN,
                Date.CAL_GREGORIAN,
                (4, 7, 1789, False, 55, 8, 1876, False),
                "Text comment",
            )

    def test_invalid_span_month(self):
        d = Date()
        with self.assertRaises(DateError):
            d.set(
                Date.QUAL_NONE,
                Date.MOD_SPAN,
                Date.CAL_GREGORIAN,
                (4, 7, 1789, False, 5, 88, 1876, False),
                "Text comment",
            )


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2007  Donald N. Allingham, Martin Hawlisch
# Copyright (C) 2009 Douglas S. Blank
# Copyright (C) 2012 Benny Malengier
# Copyright (C) 2014 Bastien Jacquet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

## Based on the paper:
##   http://www.cs.utah.edu/~draperg/research/fanchart/draperg_FHT08.pdf
## and the applet:
##   http://www.cs.utah.edu/~draperg/research/fanchart/demo/

## Found by redwood:
## http://www.gramps-project.org/bugs/view.php?id=2611

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import math
import cairo

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..utils import hex_to_rgb
from .fanchart import FanChartWidget, PIXELS_PER_GENERATION, BORDER_EDGE_WIDTH
from .fanchartdesc import (
    FanChartBaseWidget,
    FanChartDescWidget,
    FanChartGrampsGUI,
    NORMAL,
    EXPANDED,
    COLLAPSED,
    TRANSLATE_PX,
    CHILDRING_WIDTH,
    BACKGROUND_GRAD_GEN,
    BACKGROUND_GRAD_AGE,
    BACKGROUND_GRAD_PERIOD,
    FORM_CIRCLE,
    TYPE_BOX_NORMAL,
    TYPE_BOX_FAMILY,
)

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

PIXELS_PER_GENPERSON_RATIO = 0.55  # ratio of generation radius for person
# (rest for partner)
PIXELS_PER_GEN_SMALL = 80
PIXELS_PER_GEN_LARGE = 160
N_GEN_SMALL = 4
PIXELS_PER_GENFAMILY = 25  # size of radius for family
PIXELS_PER_RECLAIM = 4  # size of the radius of pixels taken from family
# to reclaim space
PIXELS_PARTNER_GAP = 0  # Padding between someone and his partner
PIXELS_CHILDREN_GAP = 5  # Padding between generations
PARENTRING_WIDTH = 12  # width of the parent ring inside the person

ANGLE_CHEQUI = 0  # Algorithm with homogeneous children distribution
ANGLE_WEIGHT = 1  # Algorithm for angle computation based on nr of descendants

TYPE_ASCENDANCE = 0
TYPE_DESCENDANCE = 1

# -------------------------------------------------------------------------
#
# FanChart2WayWidget
#
# -------------------------------------------------------------------------


class FanChart2WayWidget(FanChartWidget, FanChartDescWidget):
    """
    Interactive Fan Chart Widget.
    """

    CENTER = 50  # we require a larger center

    def __init__(self, dbstate, uistate, callback_popup=None):
        """
        Fan Chart Widget. Handles visualization of data in self.data.
        See main() of FanChartGramplet for example of model format.
        """
        self.gen2people = {}
        self.gen2fam = {}
        self.rootangle_rad_desc = [math.radians(275), math.radians(275 + 170)]
        self.rootangle_rad_asc = [math.radians(90), math.radians(270)]
        self.data = {}
        self.set_values(
            None,
            6,
            5,
            True,
            True,
            BACKGROUND_GRAD_GEN,
            True,
            "Sans",
            "#0000FF",
            "#FF0000",
            None,
            0.5,
            ANGLE_WEIGHT,
            "#888a85",
            False,
        )
        FanChartBaseWidget.__init__(self, dbstate, uistate, callback_popup)

    def reset(self):
        """
        Reset the fan chart. This should trigger computation of all data
        structures needed
        """
        self.cache_fontcolor = {}

        # fill the data structure
        self._fill_data_structures()

        # prepare the colors for the boxes
        self.prepare_background_box(self.generations_asc + self.generations_desc - 1)

    def set_values(
        self,
        root_person_handle,
        maxgen_asc,
        maxgen_desc,
        flipupsidedownname,
        twolinename,
        background,
        background_gradient,
        fontdescr,
        grad_start,
        grad_end,
        filtr,
        alpha_filter,
        angle_algo,
        dupcolor,
        showid,
    ):
        """
        Reset the values to be used:

        :param root_person_handle: person to show
        :param maxgen_asc: maximum of ascendant generations to show
        :param maxgen_desc: maximum of descendant generations to show
        :param flipupsidedownname: flip name on the left of the fanchart
                                   for the display of person's name
        :param background: config setting of which background procedure to use
        :type background: int
        :param background_gradient: option to add an overall gradient
                                    for distinguishing Asc/Desc
        :param fontdescr: string describing the font to use
        :param grad_start: colors to use for background procedure
        :param grad_end: colors to use for background procedure
        :param filtr: the person filter to apply to the people in the chart
        :param alpha_filter: the alpha transparency value (0-1) to apply to
                             filtered out data
        :param angle_algo: alorithm to use to calculate the sizes of the boxes
        :param dupcolor: color to use for people or families that occur a second
                         or more time
        :param showid: option to show the gramps id or not
        """
        self.rootpersonh = root_person_handle
        self.generations_asc = maxgen_asc
        self.generations_desc = maxgen_desc
        self.background = background
        self.background_gradient = background_gradient
        self.fontdescr = fontdescr
        self.grad_start = grad_start
        self.grad_end = grad_end
        self.filter = filtr
        self.form = FORM_CIRCLE
        self.alpha_filter = alpha_filter
        self.anglealgo = angle_algo
        self.dupcolor = hex_to_rgb(dupcolor)
        self.childring = False
        self.flipupsidedownname = flipupsidedownname
        self.twolinename = twolinename
        self.showid = showid

    def set_generations(self):
        """
        Set the generations to max, and fill data structures with initial data.
        """
        self.rootangle_rad_desc = [math.radians(275), math.radians(275 + 170)]
        self.rootangle_rad_asc = [math.radians(90), math.radians(270)]

        self.handle2desc = {}
        self.famhandle2desc = {}
        self.handle2fam = {}
        self.gen2people = {}
        self.gen2fam = {}
        # no center person
        self.gen2people[0] = [(None, False, 0, 2 * math.pi, 0, 0, [], NORMAL)]
        self.gen2fam[0] = []  # no families
        for i in range(1, self.generations_desc):
            self.gen2fam[i] = []
            self.gen2people[i] = []
        # indication of more children
        self.gen2people[self.generations_desc] = []

        # Ascendance part
        self.angle = {}
        self.data = {}
        for i in range(self.generations_asc):
            # name, person, parents?, children?
            self.data[i] = [(None,) * 4] * 2**i
            self.angle[i] = []
            angle = self.rootangle_rad_asc[0]
            portion = (
                1 / (2**i) * (self.rootangle_rad_asc[1] - self.rootangle_rad_asc[0])
            )
            for dummy_count in self.data[i]:
                # start, stop, state
                self.angle[i].append([angle, angle + portion, NORMAL])
                angle += portion

    def _fill_data_structures(self):
        """
        Initialize the data structures
        """
        self.set_generations()
        if not self.rootpersonh:
            return
        person = self.dbstate.db.get_person_from_handle(self.rootpersonh)
        if not person:
            # nothing to do, just return
            return

        # Descendance part
        # person, duplicate or not, start angle, slice size,
        #                   text, parent pos in fam, nrfam, userdata, status
        self.gen2people[0] = [[person, False, 0, 2 * math.pi, 0, 0, [], NORMAL]]
        self.handle2desc[self.rootpersonh] = 0
        # recursively fill in the datastructures:
        nrdesc = self._rec_fill_data(0, person, 0, self.generations_desc)
        self.handle2desc[person.handle] += nrdesc
        self._compute_angles(*self.rootangle_rad_desc)

        # Ascendance part
        parents = self._have_parents(person)
        child = self._have_children(person)
        # Ascendance data structure is the person object, parents, child and
        # list for userdata which we might fill in later.
        self.data[0][0] = (person, parents, child, [])
        for current in range(1, self.generations_asc):
            parent = 0
            # name, person, parents, children
            for pers, dummy_q, dummy_c, dummy_d in self.data[current - 1]:
                # Get father's and mother's details:
                for person in [
                    self._get_parent(pers, True),
                    self._get_parent(pers, False),
                ]:
                    if current == self.generations_asc - 1:
                        parents = self._have_parents(person)
                    else:
                        parents = None
                    self.data[current][parent] = (person, parents, None, [])
                    if person is None:
                        # start,stop,male/right,state
                        self.angle[current][parent][2] = COLLAPSED
                    parent += 1

    def nrgen_desc(self):
        """
        compute the number of generations present
        """
        for gen in range(self.generations_desc - 1, 0, -1):
            if self.gen2people[gen]:
                return gen + 1
        return 1

    def nrgen_asc(self):
        """
        compute the number of generations present
        """
        for generation in range(self.generations_asc - 1, 0, -1):
            for idx, (person, *rest) in enumerate(self.data[generation]):
                if person:
                    return generation
        return 1

    def maxradius_asc(self, generation):
        """
        Compute the current half radius of the ascendant circle
        """
        (dummy_radiusin, radius_asc) = self.get_radiusinout_for_gen_asc(generation)
        return radius_asc + BORDER_EDGE_WIDTH

    def maxradius_desc(self, generation):
        """
        Compute the current radius of the descendant circle
        """
        (
            dummy_radiusin_pers,
            dummy_radiusout_pers,
            dummy_radiusin_partner,
            radius_desc,
        ) = self.get_radiusinout_for_gen_pair(generation - 1)
        return radius_desc + BORDER_EDGE_WIDTH

    def halfdist(self):
        """
        Compute the current max half radius of the circle
        """
        return max(
            self.maxradius_desc(self.nrgen_desc()), self.maxradius_asc(self.nrgen_asc())
        )

    def get_radiusinout_for_gen_desc(self, generation):
        """
        Get the in and out radius for descendant generation
        (starting with center pers = 0)
        """
        radius_first_gen = (
            self.CENTER - (1 - PIXELS_PER_GENPERSON_RATIO) * PIXELS_PER_GEN_SMALL
        )
        if generation < N_GEN_SMALL:
            radius_start = PIXELS_PER_GEN_SMALL * generation + radius_first_gen
            return (radius_start, radius_start + PIXELS_PER_GEN_SMALL)
        else:
            radius_start = (
                PIXELS_PER_GEN_SMALL * N_GEN_SMALL
                + PIXELS_PER_GEN_LARGE * (generation - N_GEN_SMALL)
                + radius_first_gen
            )
            return (radius_start, radius_start + PIXELS_PER_GEN_LARGE)

    def get_radiusinout_for_gen_asc(self, generation):
        """
        Get the in and out radius for ascendant generation
        (starting with center pers = 0)
        """
        dummy_radiusin, radius_first_gen = self.get_radiusinout_for_gen_desc(0)
        outerradius = generation * PIXELS_PER_GENERATION + radius_first_gen
        innerradius = (generation - 1) * PIXELS_PER_GENERATION + radius_first_gen
        if generation == 0:
            innerradius = CHILDRING_WIDTH + TRANSLATE_PX
        return (innerradius, outerradius)

    def get_radiusinout_for_gen_pair(self, generation):
        """
        Get the in and out radius for descendant generation pair
        (starting with center pers = 0)
        :return: (radiusin_pers, radiusout_pers,
                  radiusin_partner, radiusout_partner)
        """
        radiusin, radiusout = self.get_radiusinout_for_gen_desc(generation)
        radius_spread = radiusout - radiusin - PIXELS_CHILDREN_GAP - PIXELS_PARTNER_GAP

        radiusin_pers = radiusin + PIXELS_CHILDREN_GAP
        radiusout_pers = radiusin_pers + PIXELS_PER_GENPERSON_RATIO * radius_spread
        radiusin_partner = radiusout_pers + PIXELS_PARTNER_GAP
        radiusout_partner = radiusout
        return (radiusin_pers, radiusout_pers, radiusin_partner, radiusout_partner)

    def people_generator(self):
        """
        a generator over all people outside of the core person
        """
        for generation in range(self.generations_desc):
            for data in self.gen2people[generation]:
                yield (data[0], data[6])
        for generation in range(self.generations_desc):
            for data in self.gen2fam[generation]:
                yield (data[7], data[6])
        for generation in range(self.generations_asc):
            for person, dummy_parents, dummy_child, userdata in self.data[generation]:
                yield person, userdata

    def innerpeople_generator(self):
        """
        a generator over all people inside of the core person
        """
        if False:
            yield

    def draw_background(self, ctx):
        """
        Draw the background
        """
        ctx.save()

        ctx.rotate(math.radians(self.rotate_value))
        delta = (self.rootangle_rad_asc[0] - self.rootangle_rad_desc[1]) / 2.0 % math.pi

        ctx.move_to(0, 0)
        radius_gradient_asc = 1.5 * self.maxradius_asc(self.generations_asc)
        gradient_asc = cairo.RadialGradient(
            0, 0, self.CENTER, 0, 0, radius_gradient_asc
        )
        color = hex_to_rgb(self.grad_end)
        gradient_asc.add_color_stop_rgba(
            0.0, color[0] / 255, color[1] / 255, color[2] / 255, 0.5
        )
        gradient_asc.add_color_stop_rgba(1.0, 1, 1, 1, 0.0)
        start_rad, stop_rad = (
            self.rootangle_rad_asc[0] - delta,
            self.rootangle_rad_asc[1] + delta,
        )
        ctx.set_source(gradient_asc)
        ctx.arc(0, 0, radius_gradient_asc, start_rad, stop_rad)
        ctx.fill()

        ctx.move_to(0, 0)
        radius_gradient_desc = 1.5 * self.maxradius_desc(self.generations_desc)
        gradient_desc = cairo.RadialGradient(
            0, 0, self.CENTER, 0, 0, radius_gradient_desc
        )
        color = hex_to_rgb(self.grad_start)
        gradient_desc.add_color_stop_rgba(
            0.0, color[0] / 255, color[1] / 255, color[2] / 255, 0.5
        )
        gradient_desc.add_color_stop_rgba(1.0, 1, 1, 1, 0.0)
        start_rad, stop_rad = (
            self.rootangle_rad_desc[0] - delta,
            self.rootangle_rad_desc[1] + delta,
        )
        ctx.set_source(gradient_desc)
        ctx.arc(0, 0, radius_gradient_desc, start_rad, stop_rad)
        ctx.fill()
        ctx.restore()

    def draw(self, ctx=None, scale=1.0):
        """
        The main method to do the drawing.
        If ctx is given, we assume we draw draw raw on the cairo context ctx
        To draw in GTK3 and use the allocation, set ctx=None.
        Note: when drawing for display, to counter a Gtk issue with scrolling
        or resizing the drawing window, we draw on a surface, then copy to the
        drawing context when the Gtk 'draw' signal arrives.
        """
        # first do size request of what we will need
        halfdist = self.halfdist()
        if not ctx:  # Display
            size_w = size_h = 2 * halfdist

            size_w_a = self.get_allocated_width()
            size_h_a = self.get_allocated_height()
            self.set_size_request(max(size_w, size_w_a), max(size_h, size_h_a))
            size_w = self.get_allocated_width()
            size_h = self.get_allocated_height()
            scale_factor = self.get_scale_factor()
            self.surface = cairo.ImageSurface(
                cairo.FORMAT_ARGB32, size_w * scale_factor, size_h * scale_factor
            )
            self.surface.set_device_scale(scale_factor, scale_factor)
            ctx = cairo.Context(self.surface)
            self.center_xy = self.center_xy_from_delta()
            ctx.translate(*self.center_xy)
        else:  # printing
            self.center_xy = halfdist, halfdist
            ctx.scale(scale, scale)
            ctx.translate(halfdist, halfdist)

        ctx.save()
        # Draw background
        if self.background_gradient:
            self.draw_background(ctx)
        # Draw center person:
        (
            person,
            dup,
            start,
            portion,
            dummy_parentfampos,
            dummy_nrfam,
            userdata,
            status,
        ) = self.gen2people[0][0]
        if not person:
            return
        gen_remapped = self.generations_desc - 1  # remapped generation
        if gen_remapped == 0:
            # remapped generation
            gen_remapped = self.generations_desc + self.generations_asc - 1
        (
            radiusin_pers,
            radiusout_pers,
            radiusin_partner,
            radiusout_partner,
        ) = self.get_radiusinout_for_gen_pair(0)
        radiusin = TRANSLATE_PX
        radiusout = radiusout_pers
        self.draw_person(
            ctx,
            person,
            radiusin,
            radiusout,
            math.pi / 2,
            math.pi / 2 + 2 * math.pi,
            gen_remapped,
            False,
            userdata,
            is_central_person=True,
        )
        # draw center to move chart
        ctx.set_source_rgb(0, 0, 0)  # black
        ctx.move_to(TRANSLATE_PX, 0)
        ctx.arc(0, 0, TRANSLATE_PX, 0, 2 * math.pi)
        ctx.fill()

        ctx.rotate(math.radians(self.rotate_value))
        # Ascendance
        for generation in range(self.generations_asc - 1, 0, -1):
            for idx, (person, parents, dummy_child, userdata) in enumerate(
                self.data[generation]
            ):
                if person:
                    start, stop, state = self.angle[generation][idx]
                    if state in [NORMAL, EXPANDED]:
                        fct = self.get_radiusinout_for_gen_asc
                        radiusin, radiusout = fct(generation)
                        dup = False
                        # remapped generation
                        gen_remapped = generation + self.generations_desc - 1
                        indicator = generation == self.generations_asc - 1 and parents
                        self.draw_person(
                            ctx,
                            person,
                            radiusin,
                            radiusout,
                            start,
                            stop,
                            gen_remapped,
                            dup,
                            userdata,
                            thick=(state == EXPANDED),
                            has_moregen_indicator=indicator,
                        )

        # Descendance
        for gen in range(self.generations_desc):
            (
                radiusin_pers,
                radiusout_pers,
                radiusin_partner,
                radiusout_partner,
            ) = self.get_radiusinout_for_gen_pair(gen)
            gen_remapped = self.generations_desc - gen - 1
            if gen_remapped == 0:
                # remapped generation
                gen_remapped = self.generations_desc + self.generations_asc - 1
            if gen > 0:
                for pdata in self.gen2people[gen]:
                    # person, duplicate or not, start angle, slice size,
                    #             parent pos in fam, nrfam, userdata, status
                    (
                        pers,
                        dup,
                        start,
                        portion,
                        dummy_pospar,
                        dummy_nrfam,
                        userdata,
                        status,
                    ) = pdata
                    if status != COLLAPSED:
                        self.draw_person(
                            ctx,
                            pers,
                            radiusin_pers,
                            radiusout_pers,
                            start,
                            start + portion,
                            gen_remapped,
                            dup,
                            userdata,
                            thick=status != NORMAL,
                        )
            # if gen < self.generations_desc - 1:
            for famdata in self.gen2fam[gen]:
                # family, duplicate or not, start angle, slice size,
                #       spouse pos in gen, nrchildren, userdata, status
                (
                    fam,
                    dup,
                    start,
                    portion,
                    dummy_posfam,
                    dummy_nrchild,
                    userdata,
                    partner,
                    status,
                ) = famdata
                if status != COLLAPSED:
                    more_pers_flag = (
                        gen == self.generations_desc - 1 and fam.get_child_ref_list()
                    )
                    self.draw_person(
                        ctx,
                        partner,
                        radiusin_partner,
                        radiusout_partner,
                        start,
                        start + portion,
                        gen_remapped,
                        dup,
                        userdata,
                        thick=(status != NORMAL),
                        has_moregen_indicator=more_pers_flag,
                    )
        ctx.restore()

        if self.background in [BACKGROUND_GRAD_AGE, BACKGROUND_GRAD_PERIOD]:
            self.draw_gradient_legend(ctx, halfdist)

    def cell_address_under_cursor(self, curx, cury):
        """
        Determine the cell address in the fan under the cursor
        position x and y.
        None if outside of diagram
        """
        radius, rads, dummy_raw_rads = self.cursor_to_polar(
            curx, cury, get_raw_rads=True
        )

        if radius < TRANSLATE_PX:
            return None
        radius_parents = self.get_radiusinout_for_gen_asc(0)[1]
        if (radius < radius_parents) or self.radian_in_bounds(
            self.rootangle_rad_desc[0], rads, self.rootangle_rad_desc[1]
        ):
            cell_address = self.cell_address_under_cursor_desc(rads, radius)
            if cell_address is not None:
                return (TYPE_DESCENDANCE,) + cell_address
        elif self.radian_in_bounds(
            self.rootangle_rad_asc[0], rads, self.rootangle_rad_asc[1]
        ):
            cell_address = self.cell_address_under_cursor_asc(rads, radius)
            if cell_address and cell_address[0] == 0:
                return None  # There is a gap before first parents
            if cell_address is not None:
                return (TYPE_ASCENDANCE,) + cell_address

        return None

    def cell_address_under_cursor_desc(self, rads, radius):
        """
        Determine the cell address in the fan under the cursor
        position x and y.
        None if outside of diagram
        """
        generation, selected, btype = None, None, TYPE_BOX_NORMAL
        for gen in range(self.generations_desc):
            (
                radiusin_pers,
                radiusout_pers,
                radiusin_partner,
                radiusout_partner,
            ) = self.get_radiusinout_for_gen_pair(gen)
            if radiusin_pers <= radius <= radiusout_pers:
                generation, btype = gen, TYPE_BOX_NORMAL
                break
            if radiusin_partner <= radius <= radiusout_partner:
                generation, btype = gen, TYPE_BOX_FAMILY
                break
        # find what person is in this position:
        if not (generation is None) and generation > 0:
            selected = FanChartDescWidget.personpos_at_angle_btype(
                self, generation, rads, btype
            )

        if generation is None or selected is None:
            return None

        return generation, selected, btype

    def cell_address_under_cursor_asc(self, rads, radius):
        """
        Determine the cell address in the fan under the cursor
        position x and y.
        None if outside of diagram
        """

        generation, selected = None, None
        for gen in range(self.generations_asc):
            radiusin, radiusout = self.get_radiusinout_for_gen_asc(gen)
            if radiusin <= radius <= radiusout:
                generation = gen
                break

        # find what person is in this position:
        if not (generation is None) and generation > 0:
            selected = FanChartWidget.personpos_at_angle(self, generation, rads)
        if generation is None or selected is None:
            return None
        return generation, selected

    def person_at(self, cell_address):
        """
        returns the person at radius_first_gen
        """
        direction = cell_address[0]
        if direction == TYPE_ASCENDANCE:
            return FanChartWidget.person_at(self, cell_address[1:])
        elif direction == TYPE_DESCENDANCE:
            return FanChartDescWidget.person_at(self, cell_address[1:])
        return None

    def family_at(self, cell_address):
        """
        returns the family at cell_address
        """
        direction = cell_address[0]
        if direction == TYPE_ASCENDANCE:
            return None
        elif direction == TYPE_DESCENDANCE:
            return FanChartDescWidget.family_at(self, cell_address[1:])
        return None

    def do_mouse_click(self):
        # no drag occured, expand or collapse the section
        self.toggle_cell_state(self._mouse_click_cell_address)
        self._mouse_click = False
        self.draw()
        self.queue_draw()

    def expand_parents(self, generation, selected, current):
        if generation >= self.generations_asc:
            return
        selected = 2 * selected
        start, stop, state = self.angle[generation][selected]
        if state in [NORMAL, EXPANDED]:
            portion = (stop - start) * 2.0
            self.angle[generation][selected] = [current, current + portion, state]
            self.expand_parents(generation + 1, selected, current)
            current += portion
        start, stop, state = self.angle[generation][selected + 1]
        if state in [NORMAL, EXPANDED]:
            portion = (stop - start) * 2.0
            self.angle[generation][selected + 1] = [current, current + portion, state]
            self.expand_parents(generation + 1, selected + 1, current)

    def show_parents(self, generation, selected, angle, portion):
        if generation >= self.generations_asc:
            return
        selected *= 2
        self.angle[generation][selected][0] = angle
        self.angle[generation][selected][1] = angle + portion
        self.angle[generation][selected][2] = NORMAL
        self.show_parents(generation + 1, selected, angle, portion / 2.0)
        self.angle[generation][selected + 1][0] = angle + portion
        self.angle[generation][selected + 1][1] = angle + portion + portion
        self.angle[generation][selected + 1][2] = NORMAL
        self.show_parents(generation + 1, selected + 1, angle + portion, portion / 2.0)

    def hide_parents(self, generation, selected, angle):
        if generation >= self.generations_asc:
            return
        selected = 2 * selected
        self.angle[generation][selected][0] = angle
        self.angle[generation][selected][1] = angle
        self.angle[generation][selected][2] = COLLAPSED
        self.hide_parents(generation + 1, selected, angle)
        self.angle[generation][selected + 1][0] = angle
        self.angle[generation][selected + 1][1] = angle
        self.angle[generation][selected + 1][2] = COLLAPSED
        self.hide_parents(generation + 1, selected + 1, angle)

    def shrink_parents(self, generation, selected, current):
        if generation >= self.generations_asc:
            return
        selected = 2 * selected
        start, stop, state = self.angle[generation][selected]
        if state in [NORMAL, EXPANDED]:
            portion = (stop - start) / 2.0
            self.angle[generation][selected] = [current, current + portion, state]
            self.shrink_parents(generation + 1, selected, current)
            current += portion
        start, stop, state = self.angle[generation][selected + 1]
        if state in [NORMAL, EXPANDED]:
            portion = (stop - start) / 2.0
            self.angle[generation][selected + 1] = [current, current + portion, state]
            self.shrink_parents(generation + 1, selected + 1, current)

    def toggle_cell_state(self, cell_address):
        direction = cell_address[0]
        if direction == TYPE_ASCENDANCE:
            FanChartWidget.toggle_cell_state(self, cell_address[1:])
        elif direction == TYPE_DESCENDANCE:
            FanChartDescWidget.toggle_cell_state(self, cell_address[1:])
            self._compute_angles(*self.rootangle_rad_desc)


class FanChart2WayGrampsGUI(FanChartGrampsGUI):
    """class for functions fanchart GUI elements will need in Gramps"""

    def main(self):
        """
        Fill the data structures with the active data. This initializes all
        data.
        """
        root_person_handle = self.get_active("Person")
        self.fan.set_values(
            root_person_handle,
            self.generations_asc,
            self.generations_desc,
            self.flipupsidedownname,
            self.twolinename,
            self.background,
            self.background_gradient,
            self.fonttype,
            self.grad_start,
            self.grad_end,
            self.generic_filter,
            self.alpha_filter,
            self.angle_algo,
            self.dupcolor,
            self.showid,
        )
        self.fan.reset()
        self.fan.draw()
        self.fan.queue_draw()



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    SourcePage - Source index page and individual Source pages
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from collections import defaultdict
from decimal import getcontext
from unicodedata import normalize
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import Source
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import (
    FULLCLEAR,
    html_escape,
    alphabet_navigation,
    partial_navigation,
    create_indexes_pages,
    AlphabeticIndex,
)

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb.source")
getcontext().prec = 8


#################################################
#
#    creates the Source List Page and Source Pages
#
#################################################
class SourcePages(BasePage):
    """
    This class is responsible for displaying information about the 'Source'
    database objects. It displays this information under the 'Sources'
    tab. It is told by the 'add_instances' call which 'Source's to display,
    and remembers the list of persons. A single call to 'display_pages'
    displays both the Individual List (Index) page and all the Individual
    pages.

    The base class 'BasePage' is initialised once for each page that is
    displayed.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        self.source_dict = defaultdict(set)
        self.navigation = None
        self.citationreferents = None

    def display_pages(self, the_lang, the_title):
        """
        Generate and output the pages under the Sources tab, namely the sources
        index and the individual sources pages.

        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        LOG.debug("obj_dict[Source]")
        for item in self.report.obj_dict[Source].items():
            LOG.debug("    %s", str(item))
        message = _("Creating source pages")
        progress_title = self.report.pgrs_title(the_lang)
        with self.r_user.progress(
            progress_title, message, len(self.report.obj_dict[Source]) + 1
        ) as step:
            self.sourcelistpage(
                self.report, the_lang, the_title, self.report.obj_dict[Source].keys()
            )

            index = 1
            for source_handle in self.report.obj_dict[Source]:
                step()
                index += 1
                self.sourcepage(self.report, the_lang, the_title, source_handle)

    def part_sourcelistpage(
        self,
        report,
        index_list,
        name,
        letter,
        source_handle_list,
        part=None,
        subp=None,
        partial_list=None,
    ):
        """
        Generate a partial Sources index page.
        """
        nav_name = part_name = name
        if part != 0 and subp != 0:
            part_name += str(part)
            nav_name = part_name
        if subp and subp != 0:
            part_name += subp
        output_file, sio = self.report.create_file(part_name)
        result = self.write_header(self._("Sources"))
        sourcelistpage, dummy_head, dummy_body, outerwrapper = result
        date = 0

        # begin source list division
        with Html("div", class_="content", id="Sources") as sourceslist:
            outerwrapper += sourceslist

            msg = self._(
                "This page contains an index of all the sources "
                "in the database, sorted by their title. "
                "Clicking on a source&#8217;s "
                "title will take you to that source&#8217;s page."
            )
            sourceslist += Html("p", msg, id="description")

            alpha_nav = alphabet_navigation(
                index_list,
                self.rlocale,
                current=part_name,
                rtl=self.dir,
                new_page="sources",
                ext=self.ext,
            )
            if alpha_nav is not None:
                sourceslist += alpha_nav

            if part is not None:
                # We need to create a new navigation tab for partial page index.
                partial_nav = partial_navigation(
                    partial_list,
                    self.rlocale,
                    current=part_name,
                    rtl=self.dir,
                    ext=self.ext,
                    new_page=nav_name,
                )
                if partial_nav is not None:
                    sourceslist += Html(
                        "div", style="clear: both;"
                    )  # This to align the next div to the left.
                    sourceslist += partial_nav

            # begin sourcelist table and table head
            with Html(
                "table", class_="infolist primobjlist sourcelist " + self.dir
            ) as table:
                sourceslist += table
                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                header_row = [
                    (self._("Number"), "ColumnRowLabel"),
                    (self._("Author"), "ColumnAuthor"),
                    (self._("Name", "Source Name"), "ColumnName"),
                ]

                trow.extend(
                    Html("th", label or "&nbsp;", class_=colclass, inline=True)
                    for (label, colclass) in header_row
                )

                # begin table body
                tbody = Html("tbody")
                table += tbody

                first_row = True
                for source_idx in source_handle_list:
                    source_handle, title = source_idx
                    source = self.r_db.get_source_from_handle(source_handle)

                    if first_row:
                        first_row = False
                        trow = Html("tr") + (
                            Html("td", letter, class_="ColumnRowLabel", inline=True)
                        )
                    else:
                        trow = Html("tr") + (
                            Html("td", "&nbsp;", class_="ColumnRowLabel", inline=True)
                        )
                    tbody += trow
                    trow.extend(
                        Html(
                            "td",
                            source.get_author(),
                            class_="ColumnAuthor",
                            inline=True,
                        )
                    )
                    trow.extend(
                        Html(
                            "td",
                            self.source_link(
                                source_handle,
                                source.get_title(),
                                source.get_gramps_id(),
                            ),
                            class_="ColumnName",
                        )
                    )

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(date)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(sourcelistpage, output_file, sio, 0)

    def sourcelistpage(self, report, the_lang, the_title, source_handles):
        """
        Generate and output the Sources index page.

        @param: report         -- The instance of the main report class for
                                  this report
        @param: the_lang       -- The lang to process
        @param: the_title      -- The title page related to the language
        @param: source_handles -- A list of the handles of the sources to be
                                  displayed
        """
        BasePage.__init__(self, report, the_lang, the_title)

        source_dict = {}
        # Sort the sources
        for handle in source_handles:
            source = self.r_db.get_source_from_handle(handle)
            if source is not None:
                key = source.get_title()  # + source.get_author()
                source_dict[key] = (key, handle)

        # Assemble the alphabeticIndex
        index = AlphabeticIndex(self.rlocale)
        for key, handle_list in source_dict.items():
            index.addRecord(key, handle_list)
        index_list = []
        source_handle_dict = defaultdict(list)
        max_letter_rows = defaultdict(int)
        # Extract the buckets from the index
        index.resetBucketIterator()
        while index.nextBucket():
            if index.bucketRecordCount != 0:
                letter = index.bucketLabel
                if letter not in index_list:
                    index_list.append(letter)
                bletter = normalize("NFKD", letter)[0] if len(letter) > 0 else letter
                while index.nextRecord():
                    handle_list = index.recordData
                    if handle_list:
                        source_handle_dict[bletter].append(
                            (handle_list[1], handle_list[0])
                        )
                        if bletter in max_letter_rows:
                            max_letter_rows[bletter] += len(handle_list[1])
                        else:
                            max_letter_rows[bletter] = len(handle_list[1])

        source_handle_list = list(source_handle_dict.items())
        # sort by source
        source_handle_list.sort(key=lambda x: self.rlocale.sort_key(x[0]))
        extended_handle_list = defaultdict(list)
        for bletter, hdlel in source_handle_list:
            bletter = normalize("NFKD", bletter)[0] if len(bletter) > 0 else bletter
            bletter = bletter[0] if len(bletter) > 0 else bletter
            bletter = bletter.upper() if bletter.isalpha() else "…"
            for hdle in hdlel:
                extended_handle_list[bletter].append((hdle))
        extended_handle_list = list(extended_handle_list.items())
        create_indexes_pages(
            report,
            "sources",
            index_list,
            self.part_sourcelistpage,
            extended_handle_list,
            locale=self.rlocale,
        )

    def sourcepage(self, report, the_lang, the_title, source_handle):
        """
        Generate and output an individual Source page.

        @param: report        -- The instance of the main report class
                                 for this report
        @param: the_lang      -- The lang to process
        @param: the_title     -- The title page related to the language
        @param: source_handle -- The handle of the source to be output
        """
        source = report.database.get_source_from_handle(source_handle)
        BasePage.__init__(self, report, the_lang, the_title, source.get_gramps_id())
        if not source:
            return

        self.page_title = source.get_title()

        inc_repositories = self.report.options["inc_repository"]
        self.navigation = self.report.options["navigation"]
        self.citationreferents = self.report.options["citationreferents"]

        output_file, sio = self.report.create_file(source_handle, "src")
        self.uplink = True
        result = self.write_header("%s - %s" % (self._("Sources"), self.page_title))
        sourcepage, head, dummy_body, outerwrapper = result

        ldatec = 0
        # begin source detail division
        with Html("div", class_="content", id="SourceDetail") as sourcedetail:
            outerwrapper += sourcedetail

            media_list = source.get_media_list()
            if self.create_media and media_list:
                if self.the_lang and not self.usecms:
                    fname = "/".join(["..", "css", "lightbox.css"])
                    jsname = "/".join(["..", "css", "lightbox.js"])
                else:
                    fname = "/".join(["css", "lightbox.css"])
                    jsname = "/".join(["css", "lightbox.js"])
                url = self.report.build_url_fname(fname, None, self.uplink)
                head += Html(
                    "link", href=url, type="text/css", media="screen", rel="stylesheet"
                )
                url = self.report.build_url_fname(jsname, None, self.uplink)
                head += Html("script", src=url, type="text/javascript", inline=True)
                thumbnail = self.disp_first_img_as_thumbnail(media_list, source)
                if thumbnail is not None:
                    sourcedetail += thumbnail

            # add section title
            sourcedetail += Html("h3", html_escape(source.get_title()), inline=True)

            # begin sources table
            with Html("table", class_="infolist source " + self.dir) as table:
                sourcedetail += table

                tbody = Html("tbody")
                table += tbody

                source_gid = False
                if not self.noid and self.gid:
                    source_gid = source.get_gramps_id()

                    # last modification of this source
                    ldatec = source.get_change_time()

                for label, value in [
                    (self._("Gramps ID"), source_gid),
                    (self._("Author"), source.get_author()),
                    (self._("Abbreviation"), source.get_abbreviation()),
                    (self._("Publication information"), source.get_publication_info()),
                ]:
                    if value:
                        trow = Html("tr") + (
                            Html("td", label, class_="ColumnAttribute", inline=True),
                            Html("td", value, class_="ColumnValue", inline=True),
                        )
                        tbody += trow

            # Tags
            tags = self.show_tags(source)
            if tags and self.report.inc_tags:
                trow = Html("tr") + (
                    Html("td", self._("Tags"), class_="ColumnAttribute", inline=True),
                    Html("td", tags, class_="ColumnValue", inline=True),
                )
                tbody += trow

            # Source notes
            notelist = self.display_note_list(source.get_note_list(), Source)
            if notelist is not None:
                sourcedetail += notelist

            # additional media from Source (if any?)
            if self.create_media and media_list:
                sourcemedia = self.disp_add_img_as_gallery(media_list, source)
                if sourcemedia is not None:
                    sourcedetail += sourcemedia

            # Source Data Map...
            src_data_map = self.write_srcattr(source.get_attribute_list())
            if src_data_map is not None:
                sourcedetail += src_data_map

            # Source Repository list
            if inc_repositories:
                repo_list = self.dump_repository_ref_list(source.get_reporef_list())
                if repo_list is not None:
                    sourcedetail += repo_list

            # Source references list
            ref_list = self.display_bkref_list(Source, source_handle)
            if ref_list is not None:
                sourcedetail += ref_list

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(sourcepage, output_file, sio, ldatec)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....utils.db import for_each_ancestor
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Dict, Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasCommonAncestorWith
#
# -------------------------------------------------------------------------
class HasCommonAncestorWith(Rule):
    """Rule that checks for a person that has a common ancestor with a specified person"""

    labels = [_("ID:")]
    name = _("People with a common ancestor with <person>")
    category = _("Ancestral filters")
    description = _(
        "Matches people that have a common ancestor " "with a specified person"
    )

    def prepare(self, db: Database, user):
        self.db = db
        # For each(!) person we keep track of who their ancestors
        # are, in a set(). So we only have to compute a person's
        # ancestor list once.
        # Start with filling the cache for root person (gramps_id in self.list[0])
        self.ancestor_cache: Dict[str, Set[str]] = {}
        root_person = db.get_person_from_gramps_id(self.list[0])
        if root_person:
            self.add_ancs(db, root_person)
            self.with_people = [root_person.handle]
        else:
            self.with_people = []

    def add_ancs(self, db: Database, person: Person):
        if person and person.handle not in self.ancestor_cache:
            self.ancestor_cache[person.handle] = set()
            # We are going to compare ancestors of one person with that of
            # another person; if that other person is an ancestor and itself
            # has no ancestors is must be included, this is achieved by the
            # little trick of making a person his own ancestor.
            self.ancestor_cache[person.handle].add(person.handle)
        else:
            return

        for fam_handle in person.parent_family_list:
            parentless_fam = True
            fam = db.get_family_from_handle(fam_handle)
            if fam:
                for par_handle in (fam.father_handle, fam.mother_handle):
                    if par_handle:
                        parentless_fam = False
                        par = db.get_person_from_handle(par_handle)
                        if par and par.handle not in self.ancestor_cache:
                            self.add_ancs(db, par)
                        if par:
                            self.ancestor_cache[person.handle] |= self.ancestor_cache[
                                par.handle
                            ]
                if parentless_fam:
                    self.ancestor_cache[person.handle].add(fam_handle)

    def reset(self):
        self.ancestor_cache = {}

    def has_common_ancestor(self, other: Person):
        for handle in self.with_people:
            left_and = (
                handle in self.ancestor_cache and self.ancestor_cache[handle]
            )  # type: ignore
            right_and = (
                other
                and other.handle in self.ancestor_cache
                and self.ancestor_cache[other.handle]
            )  # type: ignore
            if left_and.intersection(right_and):  # type: ignore
                return True
        return False

    def apply_to_one(self, db: Database, person: Person) -> bool:
        if person and person.handle not in self.ancestor_cache:
            self.add_ancs(db, person)

        return self.has_common_ancestor(person)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for a repo with a particular value.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....lib.repotype import RepositoryType
from .. import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Repository
from ....db import Database


_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# HasRepo
#
# -------------------------------------------------------------------------
class HasRepo(Rule):
    """
    Rule that checks for a repo with a particular value.
    """

    labels = [
        _("Name:", "repo"),
        _("Type:"),
        _("Address:"),
        _("URL:"),
    ]
    name = _("Repositories matching parameters")
    description = _("Matches Repositories with particular parameters")
    category = _("General filters")
    allow_regex = True

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.rtype = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things we only want to do once.
        """
        if self.list[1]:
            self.rtype = RepositoryType()
            self.rtype.set_from_xml_str(self.list[1])

    def apply_to_one(self, _db: Database, obj: Repository) -> bool:
        """
        Apply the rule. Return True on a match.
        """
        if not self.match_substring(0, obj.name):
            return False

        if self.rtype:
            if self.rtype.is_custom() and self.use_regex:
                if self.regex[1].search(str(obj.type)) is None:
                    return False
            elif obj.type != self.rtype:
                return False

        if self.list[2]:
            addr_match = False
            for addr in obj.address_list:
                # TODO for Arabic, should the next line's comma be translated?
                addr_text = ", ".join(addr.get_text_data_list())
                if self.match_substring(2, addr_text):
                    addr_match = True
                    break
            if not addr_match:
                return False

        if self.list[3]:
            url_match = False
            for url in obj.urls:
                # TODO for Arabic, should the next line's comma be translated?
                url_text = ", ".join(url.get_text_data_list())
                if self.match_substring(3, url_text):
                    url_match = True
                    break
            if not url_match:
                return False

        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from .backrefmodel import BackRefModel
from .backreflist import BackRefList


class PersonBackRefList(BackRefList):
    def __init__(self, dbstate, uistate, track, obj, config_key, callback=None):
        BackRefList.__init__(
            self,
            dbstate,
            uistate,
            track,
            obj,
            BackRefModel,
            config_key,
            callback=callback,
        )

    def get_icon_name(self):
        return "gramps-relation"



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2019  Matthias Kemmer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for an object with a particular attribute.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale
from ...lib.attrtype import AttributeType
from . import Rule

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib.attrbase import AttributeBase
from ...db import Database


# -------------------------------------------------------------------------
#
# HasAttributeBase
#
# -------------------------------------------------------------------------
class HasAttributeBase(Rule):
    """
    Rule that checks for an object with a particular attribute.
    """

    labels = ["Attribute:", "Value:"]
    name = "Objects with the <attribute>"
    description = "Matches objects with the given attribute of a particular value"
    category = _("General filters")
    allow_regex = True

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.attribute_type = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things that should only be done once.
        """
        if self.list[0]:
            self.attribute_type = AttributeType()
            self.attribute_type.set_from_xml_str(self.list[0])

    def apply_to_one(self, db: Database, obj: AttributeBase) -> bool:
        """
        Apply the rule. Return True if a match.
        """
        if self.attribute_type:
            for attribute in obj.attribute_list:
                name_match = attribute.type == self.attribute_type
                if name_match:
                    if self.match_substring(1, attribute.value):
                        return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The EditChildRef module provides the EditChildRef class.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import pickle

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from .editsecondary import EditSecondary
from gramps.gen.lib import NoteType
from gramps.gen.errors import WindowActiveError
from ..glade import Glade
from ..ddtargets import DdTargets
from .displaytabs import CitationEmbedList, NoteTab
from ..widgets import MonitoredDataType, PrivacyButton
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.const import URL_MANUAL_SECT1

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT1
WIKI_HELP_SEC = _("Child_Reference_Editor", "manual")

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")
_LEFT_BUTTON = 1
_RIGHT_BUTTON = 3


# -------------------------------------------------------------------------
#
# EditChildRef class
#
# -------------------------------------------------------------------------
class EditChildRef(EditSecondary):
    """
    Displays a dialog that allows the user to edit a child reference.
    """

    def __init__(self, name, dbstate, uistate, track, childref, callback):
        """
        Displays the dialog box.

        childref - The child reference that is to be edited
        """
        self.name = name
        EditSecondary.__init__(self, dbstate, uistate, track, childref, callback)

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel,
            self.top.get_object("title"),
            self.name,
            _("Child Reference Editor"),
        )
        self.setup_configs("interface.child-ref", 600, 450)

        self.ok_button = self.top.get_object("ok")
        self.edit_button = self.top.get_object("edit")
        self.name_label = self.top.get_object("name")
        self.name_label.set_text(self.name)

        # Set the drag action from the label
        self.label_event_box = self.top.get_object("name_event_box")
        self.label_event_box.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [DdTargets.PERSON_LINK.target()],
            Gdk.DragAction.COPY,
        )
        self.label_event_box.drag_source_set_icon_name("gramps-person")
        self.label_event_box.connect("drag_data_get", self.drag_data_get)

    def _setup_fields(self):
        self.frel = MonitoredDataType(
            self.top.get_object("frel"),
            self.obj.set_father_relation,
            self.obj.get_father_relation,
            self.db.readonly,
            self.db.get_child_reference_types(),
        )

        self.mrel = MonitoredDataType(
            self.top.get_object("mrel"),
            self.obj.set_mother_relation,
            self.obj.get_mother_relation,
            self.db.readonly,
            self.db.get_child_reference_types(),
        )

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_ok_button(self.ok_button, self.save)
        self.edit_button.connect("button-press-event", self.edit_child)
        self.edit_button.connect("key-press-event", self.edit_child)
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("person-update", self.person_change)
        self._add_db_signal("person-rebuild", self.close)
        self._add_db_signal("person-delete", self.check_for_close)

    def drag_data_get(self, widget, context, sel_data, info, time):
        # get the selected object, returning if not is defined
        if not self.obj.ref:
            return
        if info == DdTargets.PERSON_LINK.app_id:
            data = (DdTargets.PERSON_LINK.drag_type, id(self), self.obj.ref, 0)
            sel_data.set(DdTargets.PERSON_LINK.atom_drag_type, 8, pickle.dumps(data))

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """
        notebook = Gtk.Notebook()

        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "childref_editor_citations",
        )
        self._add_tab(notebook, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "childref_editor_notes",
            notetype=NoteType.CHILDREF,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def _post_init(self):
        self.ok_button.grab_focus()

    def build_menu_names(self, obj):
        return (_("Child Reference"), _("Child Reference Editor"))

    def edit_child(self, obj, event):
        if button_activated(event, _LEFT_BUTTON):
            from .editperson import EditPerson

            handle = self.obj.ref
            try:
                person = self.db.get_person_from_handle(handle)
                EditPerson(self.dbstate, self.uistate, self.track, person)
            except WindowActiveError:
                pass

    def person_change(self, handles):
        # check to see if the handle matches the current object
        if self.obj.ref in handles:
            p = self.dbstate.db.get_person_from_handle(self.obj.ref)
            self.name = name_displayer.display(p)
            self.name_label.set_text(self.name)

    def save(self, *obj):
        """
        Called when the OK button is pressed. Gets data from the
        form and updates the ChildRef data structure.
        """
        if self.callback:
            self.callback(self.obj)
        self.close()

    def check_for_close(self, handles):
        """
        Callback method for delete signals.
        If there is a delete signal of the primary object we are editing, the
        editor (and all child windows spawned) should be closed
        """
        if self.obj.ref in handles:
            self.close()


def button_activated(event, mouse_button):
    if (event.type == Gdk.EventType.BUTTON_PRESS and event.button == mouse_button) or (
        event.type == Gdk.EventType.KEY_PRESS and event.keyval in (_RETURN, _KP_ENTER)
    ):
        return True
    else:
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008 Brian G. Matherly
# Copyright (C) 2008      Stephane Charette <stephanecharette@gmail.com>
# Contribution 2009 by    Bob Ham <rah@bash.sh>
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2013-2014 Paul Franklin
# Copyright (C) 2015      Detlef Wolz <detlef.wolz@t-online.de>
# Copyright (C) 2018      Christophe aka khrys63
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Generate an hourglass graph using the Graphviz generator.
"""
# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import html

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.errors import ReportError
from gramps.gen.plug.menu import (
    PersonOption,
    BooleanOption,
    NumberOption,
    EnumeratedListOption,
    ColorOption,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.utils.symbols import Symbols

# ------------------------------------------------------------------------
#
# Constant options items
#
# ------------------------------------------------------------------------
_COLORS = [
    {"name": _("B&W outline"), "value": "outline"},
    {"name": _("Colored outline"), "value": "colored"},
    {"name": _("Color fill"), "value": "filled"},
]

_ARROWS = [
    {"name": _("Center -> Others"), "value": "o"},
    {"name": _("Center <- Others"), "value": "c"},
    {"name": _("Center <-> Other"), "value": "co"},
    {"name": _("Center - Other"), "value": ""},
]


# ------------------------------------------------------------------------
#
# HourGlassReport
#
# ------------------------------------------------------------------------
class HourGlassReport(Report):
    """
    An hourglass report displays ancestors and descendants of a center person.
    """

    def __init__(self, database, options, user):
        """
        Create HourGlass object that produces the report.

        name_format   - Preferred format to display names
        incl_private  - Whether to include private data
        inc_id        - Whether to include IDs.
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """
        Report.__init__(self, database, options, user)
        menu = options.menu

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)
        self.__db = self.database

        self.__used_people = []
        self.__family_father = []  # links allocated from family to father
        self.__family_mother = []  # links allocated from family to mother

        self.__node_label = {}  # labels of node for merge sosa number

        self.max_descend = menu.get_option_by_name("maxdescend").get_value()
        self.max_ascend = menu.get_option_by_name("maxascend").get_value()
        pid = menu.get_option_by_name("pid").get_value()
        self.center_person = self.__db.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        self.colorize = menu.get_option_by_name("color").get_value()
        self.colors = {
            "male": menu.get_option_by_name("colormales").get_value(),
            "female": menu.get_option_by_name("colorfemales").get_value(),
            "other": menu.get_option_by_name("colorother").get_value(),
            "unknown": menu.get_option_by_name("colorunknown").get_value(),
            "family": menu.get_option_by_name("colorfamilies").get_value(),
        }
        self.roundcorners = menu.get_option_by_name("roundcorners").get_value()

        self.ahnentafel = menu.get_option_by_name("ahnentafel").get_value()

        self.ahnentafelnum = menu.get_option_by_name("ahnentafelnum").get_value()

        self.datesymbol = menu.get_option_by_name("datesymbol").get_value()

        self.includeid = menu.get_option_by_name("inc_id").get_value()

        arrow_str = menu.get_option_by_name("arrow").get_value()
        if "o" in arrow_str:
            self.arrowheadstyle = "normal"
        else:
            self.arrowheadstyle = "none"
        if "c" in arrow_str:
            self.arrowtailstyle = "normal"
        else:
            self.arrowtailstyle = "none"

        stdoptions.run_name_format_option(self, menu)

    def write_report(self):
        """
        Generate the report.
        """
        self.add_person(self.center_person, 1)
        self.traverse_up(self.center_person, 1, 1)
        self.traverse_down(self.center_person, 1)

    def traverse_down(self, person, gen):
        """
        Recursively find the descendants of the given person.
        """
        if gen > self.max_descend:
            return
        for family_handle in person.get_family_handle_list():
            family = self.__db.get_family_from_handle(family_handle)
            self.add_family(family)
            self.doc.add_link(
                person.get_gramps_id(),
                family.get_gramps_id(),
                head=self.arrowheadstyle,
                tail=self.arrowtailstyle,
            )
            for child_ref in family.get_child_ref_list():
                child_handle = child_ref.get_reference_handle()
                if child_handle not in self.__used_people:
                    # Avoid going down paths twice when descendant cousins marry
                    self.__used_people.append(child_handle)
                    child = self.__db.get_person_from_handle(child_handle)
                    self.add_person(child, 0)
                    self.doc.add_link(
                        family.get_gramps_id(),
                        child.get_gramps_id(),
                        head=self.arrowheadstyle,
                        tail=self.arrowtailstyle,
                    )
                    self.traverse_down(child, gen + 1)

    def traverse_up(self, person, gen, sosanumber):
        """
        Recursively find the ancestors of the given person.
        """
        fathersosanumber = sosanumber * 2
        mothersosanumber = fathersosanumber + 1
        if gen > self.max_ascend:
            return
        family_handle = person.get_main_parents_family_handle()
        person_id = person.get_gramps_id()
        if family_handle:
            family = self.__db.get_family_from_handle(family_handle)
            family_id = family.get_gramps_id()
            self.add_family(family)
            self.doc.add_link(
                family_id, person_id, head=self.arrowtailstyle, tail=self.arrowheadstyle
            )
            father_id = ""
            mother_id = ""

            # create link from family to father
            father_handle = family.get_father_handle()
            if father_handle and family_handle not in self.__family_father:
                # allocate only one father per family
                self.__family_father.append(family_handle)
                father = self.__db.get_person_from_handle(father_handle)
                self.add_person(father, fathersosanumber)
                father_id = father.get_gramps_id()
                self.doc.add_link(
                    father.get_gramps_id(),
                    family_id,
                    head=self.arrowtailstyle,
                    tail=self.arrowheadstyle,
                )
                # update node with its father node id
                self.__node_label[person_id] = [
                    self.__node_label[person_id][0],
                    father_id,
                    self.__node_label[person_id][2],
                ]
                # no need to go up if he is a father in another family
                if father_handle not in self.__used_people:
                    self.__used_people.append(father_handle)
                    self.traverse_up(father, gen + 1, fathersosanumber)
            elif family_handle in self.__family_father and self.ahnentafelnum:
                self.rewrite_sosa_number(
                    self.__db.get_person_from_handle(father_handle).get_gramps_id(),
                    fathersosanumber,
                )

            # create link from family to mother
            mother_handle = family.get_mother_handle()
            if mother_handle and family_handle not in self.__family_mother:
                # allocate only one mother per family
                self.__family_mother.append(family_handle)
                mother = self.__db.get_person_from_handle(mother_handle)
                self.add_person(mother, mothersosanumber)
                mother_id = mother.get_gramps_id()
                self.doc.add_link(
                    mother.get_gramps_id(),
                    family_id,
                    head=self.arrowtailstyle,
                    tail=self.arrowheadstyle,
                )
                # update node with its mother node id
                self.__node_label[person_id] = [
                    self.__node_label[person_id][0],
                    self.__node_label[person_id][1],
                    mother_id,
                ]
                # no need to go up if she is a mother in another family
                if mother_handle not in self.__used_people:
                    self.__used_people.append(mother_handle)
                    self.traverse_up(mother, gen + 1, mothersosanumber)
            elif family_handle in self.__family_mother and self.ahnentafelnum:
                self.rewrite_sosa_number(
                    self.__db.get_person_from_handle(mother_handle).get_gramps_id(),
                    mothersosanumber,
                )

            if (
                self.ahnentafel
                and mother_handle
                and father_handle
                and father_id != ""
                and mother_id != ""
            ):
                self.doc.add_link(father_id, mother_id, style="invis")
                self.doc.add_samerank(father_id, mother_id)

    def rewrite_sosa_number(self, pid, sosanumber):
        """
        Rewrite the Sosa number of a node for multiple sosa member in the tree.
        """
        self.__node_label[pid][0] += " - #%s" % (sosanumber)
        self.doc.rewrite_label(pid, self.__node_label[pid][0])

        # Recursively rewrite for all ancestors
        if self.__node_label[pid][1] != "":
            self.rewrite_sosa_number(self.__node_label[pid][1], sosanumber * 2)
        if self.__node_label[pid][2] != "":
            self.rewrite_sosa_number(self.__node_label[pid][2], sosanumber * 2 + 1)

    def add_person(self, person, sosanumber):
        """
        Add a person to the Graph. The node id will be the person's gramps id.
        """
        p_id = person.get_gramps_id()
        name = self._name_display.display(person)
        name = html.escape(name)
        if self.datesymbol:
            symbols = Symbols()

        birth_evt = get_birth_or_fallback(self.__db, person)
        if birth_evt:
            birth = self.get_date(birth_evt.get_date_object())
            if self.datesymbol:
                birth = (
                    symbols.get_symbol_for_string(symbols.SYMBOL_BIRTH) + " %s" % birth
                )
        else:
            birth = ""

        death_evt = get_death_or_fallback(self.__db, person)
        if death_evt:
            death = self.get_date(death_evt.get_date_object())
            if self.datesymbol:
                death = (
                    symbols.get_death_symbol_fallback(symbols.DEATH_SYMBOL_LATIN_CROSS)
                    + " %s" % death
                )
        else:
            death = ""

        if death and self.datesymbol == False:
            death = " – %s" % death
        elif death and birth:
            death = " %s" % death

        if self.includeid == 0:  # no ID
            label = "%s \\n(%s%s)" % (name, birth, death)
        elif self.includeid == 1:  # same line
            label = "%s (%s)\\n(%s%s)" % (name, p_id, birth, death)
        elif self.includeid == 2:  # own line
            label = "%s \\n(%s%s)\\n(%s)" % (name, birth, death, p_id)

        if self.ahnentafelnum and sosanumber != 0:
            label += "\\n #%s" % (sosanumber)

        label = label.replace('"', '\\"')

        (shape, style, color, fill) = self.get_gender_style(person)
        self.doc.add_node(p_id, label, shape, color, style, fill)

        # save node with them label, father node id, mother node id and sosanumber
        self.__node_label[p_id] = [label, "", ""]

    def add_family(self, family):
        """
        Add a family to the Graph. The node id will be the family's gramps id.
        """
        family_id = family.get_gramps_id()
        if self.datesymbol:
            symbols = Symbols()
        label = ""
        marriage = utils.find_marriage(self.__db, family)
        if marriage:
            label = self.get_date(marriage.get_date_object())
        if self.includeid == 1 and label:  # same line
            label = "%s (%s)" % (label, family_id)
        elif self.includeid == 1 and not label:
            label = "(%s)" % family_id
        elif self.includeid == 2 and label:  # own line
            label = "%s\\n(%s)" % (label, family_id)
        elif self.includeid == 2 and not label:
            label = "(%s)" % family_id
        if self.datesymbol and label:
            label = (
                symbols.get_symbol_for_string(symbols.SYMBOL_MARRIAGE) + " %s" % label
            )

        color = ""
        fill = ""
        style = "solid"
        if self.colorize == "colored":
            color = self.colors["family"]
        elif self.colorize == "filled":
            fill = self.colors["family"]
            style = "filled"
        self.doc.add_node(family_id, label, "ellipse", color, style, fill)

    def get_gender_style(self, person):
        "return gender specific person style"
        gender = person.get_gender()
        shape = "box"
        style = "solid"
        color = ""
        fill = ""

        if gender == person.FEMALE and self.roundcorners:
            style = "rounded"
        elif gender == person.UNKNOWN:
            shape = "hexagon"

        if self.colorize == "colored":
            if gender == person.MALE:
                color = self.colors["male"]
            elif gender == person.FEMALE:
                color = self.colors["female"]
            elif gender == person.OTHER:
                color = self.colors["other"]
            else:
                color = self.colors["unknown"]
        elif self.colorize == "filled":
            style += ",filled"
            if gender == person.MALE:
                fill = self.colors["male"]
            elif gender == person.FEMALE:
                fill = self.colors["female"]
            elif gender == person.OTHER:
                fill = self.colors["other"]
            else:
                fill = self.colors["unknown"]
        return (shape, style, color, fill)

    def get_date(self, date):
        """return a formatted date"""
        return html.escape(self._get_date(date))


# ------------------------------------------------------------------------
#
# HourGlassOptions
#
# ------------------------------------------------------------------------
class HourGlassOptions(MenuReportOptions):
    """
    Defines options for the HourGlass report.
    """

    def __init__(self, name, dbase):
        MenuReportOptions.__init__(self, name, dbase)

    def add_menu_options(self, menu):
        """
        Create all the menu options for this report.
        """
        category_name = _("Report Options")

        pid = PersonOption(_("Center Person"))
        pid.set_help(_("The Center person for the graph"))
        menu.add_option(category_name, "pid", pid)

        max_gen_d = NumberOption(_("Max Descendant Generations"), 10, 0, 15)
        max_gen_d.set_help(
            _("The number of generations of descendants to " "include in the graph")
        )
        menu.add_option(category_name, "maxdescend", max_gen_d)

        max_gen_a = NumberOption(_("Max Ancestor Generations"), 10, 0, 15)
        max_gen_a.set_help(
            _("The number of generations of ancestors to " "include in the graph")
        )
        menu.add_option(category_name, "maxascend", max_gen_a)

        arrow = EnumeratedListOption(_("Arrowhead direction"), "o")
        for i in range(0, len(_ARROWS)):
            arrow.add_item(_ARROWS[i]["value"], _ARROWS[i]["name"])
        arrow.set_help(_("Choose the direction that the arrows point."))
        menu.add_option(category_name, "arrow", arrow)

        color = EnumeratedListOption(_("Graph coloring"), "filled")
        for i in range(0, len(_COLORS)):
            color.add_item(_COLORS[i]["value"], _COLORS[i]["name"])
        color.set_help(
            _(
                "Males will be shown with blue, females "
                "with red.  If the sex of an individual "
                "is unknown it will be shown with gray."
            )
        )
        menu.add_option(category_name, "color", color)

        roundedcorners = BooleanOption(_("Use rounded corners"), False)  # 2180
        roundedcorners.set_help(
            _("Use rounded corners to differentiate between women and men.")
        )
        menu.add_option(category_name, "roundcorners", roundedcorners)

        stdoptions.add_gramps_id_option(menu, category_name, ownline=True)

        ################################
        category_name = _("Report Options (2)")
        ################################

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

        ################################
        category_name = _("Graph Style")
        ################################

        color_males = ColorOption(_("Males"), "#e0e0ff")
        color_males.set_help(_("The color to use to display men."))
        menu.add_option(category_name, "colormales", color_males)

        color_females = ColorOption(_("Females"), "#ffe0e0")
        color_females.set_help(_("The color to use to display women."))
        menu.add_option(category_name, "colorfemales", color_females)

        color_other = ColorOption(_("Other"), "#94ef9e")
        color_other.set_help(
            _("The color to use to display people who are " "neither men nor women.")
        )
        menu.add_option(category_name, "colorother", color_other)

        color_unknown = ColorOption(_("Unknown"), "#e0e0e0")
        color_unknown.set_help(_("The color to use " "when the gender is unknown."))
        menu.add_option(category_name, "colorunknown", color_unknown)

        color_family = ColorOption(_("Families"), "#ffffe0")
        color_family.set_help(_("The color to use to display families."))
        menu.add_option(category_name, "colorfamilies", color_family)

        ahnentafelorder = BooleanOption(_("Force Ahnentafel order"), False)  # 10826
        ahnentafelorder.set_help(
            _(
                "Force Sosa / Sosa-Stradonitz / Ahnentafel layout order for all ancestors, so that fathers are always on the left branch and mothers are on the right branch."
            )
        )
        menu.add_option(category_name, "ahnentafel", ahnentafelorder)

        ahnentafelnumvisible = BooleanOption(
            _("Ahnentafel number visible"), False
        )  # 10826
        ahnentafelnumvisible.set_help(
            _("Show Sosa / Sosa-Stradonitz / Ahnentafel number.")
        )
        menu.add_option(category_name, "ahnentafelnum", ahnentafelnumvisible)

        datesymbol = BooleanOption(_("Show symbols for events"), False)
        datesymbol.set_help(
            _("Show genealogical symbols for birth, marriage and death events.")
        )
        menu.add_option(category_name, "datesymbol", datesymbol)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017,2024       Nick Hall
# Copyright (C) 2024-2025       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------

from __future__ import annotations
import json
import orjson

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------

import gramps.gen.lib as lib


NoneType = type(None)


class DataDict(dict):
    """
    A wrapper around a data dict that also provides an
    object interface.
    """

    def __init__(self, data=None):
        """
        Wrap a data dict (raw data) or object
        with an attribute API. If data is an
        object, we use it to get the attributes.
        """
        if isinstance(data, dict):
            super().__init__(data)
        else:
            super().__init__()
            if data:
                # Data is actually a Gramps object
                self.update(object_to_data(data))

    def __str__(self):
        if "_object" not in self:
            self["_object"] = data_to_object(self)
        return str(self["_object"])

    def __getattr__(self, key):
        if key.startswith("_"):
            raise AttributeError(
                "this method cannot be used to access hidden attributes"
            )

        if key in self:
            value = self[key]
            if isinstance(value, dict):
                return DataDict(value)
            elif isinstance(value, list):
                return DataList(value)
            else:
                return value
        else:
            # Some method or attribute not available
            # otherwise. Cache it:
            self["_object"] = data_to_object(self)
            # And return the attr from the object:
            return getattr(self["_object"], key)


class DataList(list):
    """
    A wrapper around a data list.
    """

    def __getitem__(self, position):
        value = super().__getitem__(position)
        if isinstance(value, dict):
            return DataDict(value)
        elif isinstance(value, list):
            return DataList(value)
        else:
            return value

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        if self.index >= len(self):
            raise StopIteration

        result = self[self.index]
        self.index += 1
        return result

    def __add__(self, value):
        return DataList([x for x in self] + [x for x in value])

    def __radd__(self, value):
        return DataList([x for x in self] + [x for x in value])


def remove_object(data):
    """
    Remove _object, if any, from dicts.
    """
    if isinstance(data, dict):
        return {
            key: remove_object(value) for key, value in data.items() if key != "_object"
        }
    elif isinstance(data, (list, tuple)):
        return [remove_object(item) for item in data]
    else:
        return data


def convert_state_to_object(obj_dict):
    _class = obj_dict.pop("_class")
    cls = lib.__dict__[_class]
    obj = cls.__new__(cls)
    obj.set_object_state(obj_dict)
    return obj


def convert_object_to_state(obj):
    return obj.get_object_state()


def string_to_data(string: str | bytes) -> DataDict:
    """
    Convert a JSON string into its data representation.
    """
    return DataDict(orjson.loads(string))


def string_to_dict(string: str | bytes) -> dict:
    """
    Convert a JSON string into its dict representation.
    """
    return orjson.loads(string)


def dict_to_string(dict: dict) -> str:
    """
    Convert a dict into its JSON string representation.
    """
    return orjson.dumps(dict).decode()


def object_to_dict(obj):
    """
    Convert any Gramps lib object into a dict representation.
    """
    if isinstance(obj, (float, int, str, NoneType)):
        return obj

    elif isinstance(obj, (list, tuple)):
        return [object_to_dict(item) for item in obj]

    state = obj.get_object_state()
    return {k: object_to_dict(v) for k, v in state.items()}


def object_to_data(obj):
    """
    Convert any Gramps lib object or other value into
    its dict or list representation or value, respectively.
    """
    if isinstance(obj, (float, int, str, NoneType)):
        return obj

    elif isinstance(obj, (list, tuple)):
        return DataList([object_to_data(item) for item in obj])

    state = obj.get_object_state()
    data = {k: object_to_data(v) for k, v in state.items()}
    # Stash the object in case we need it later:
    data["_object"] = obj
    return DataDict(data)


def data_to_object(data):
    """
    Convert any object dict representation to a Gramps lib object
    or other value.
    """
    if isinstance(data, dict):
        if "_object" in data:
            return data["_object"]

        data = {k: data_to_object(v) for k, v in data.items()}
        return convert_state_to_object(data)

    elif isinstance(data, (list, tuple)):
        return [data_to_object(item) for item in data]

    return data


def object_to_string(obj: object) -> str:
    """
    Convert any Gramps object into a JSON string/bytes.
    """
    return orjson.dumps(obj, default=convert_object_to_state).decode()


def data_to_string(data: DataDict) -> str:
    """
    Convert a DataDict into a string.
    """
    return orjson.dumps(remove_object(data)).decode()


def string_to_object(string: str | bytes):
    """
    Convert a JSON string/bytes into a Gramps lib object.
    """
    return data_to_object(orjson.loads(string))
